from __future__ import annotations

import logging
import pathlib
import re
import sys
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from pgtoolkit.conf import parse as parse_pgconf

from pglift import exceptions, instances, postgresql
from pglift.ctx import Context
from pglift.models import interface, system
from pglift.postgresql import (
    ctl,
    site_configure_install,
    site_configure_installed,
    site_configure_uninstall,
    systemd_unit_templates,
)
from pglift.settings import Settings, _postgresql
from pglift.types import ConfigChanges, Status


def test_site_configure(settings: Settings) -> None:
    assert not site_configure_installed(settings)
    assert not settings.postgresql.logpath.exists()
    site_configure_install(settings)
    assert site_configure_installed(settings)
    assert settings.postgresql.logpath.exists()
    site_configure_uninstall(settings)
    assert not settings.postgresql.logpath.exists()


def test_initdb_dirty(
    pg_version: str, settings: Settings, ctx: Context, monkeypatch: pytest.MonkeyPatch
) -> None:
    manifest = interface.Instance(name="dirty", version=pg_version)
    i = system.BaseInstance("dirty", pg_version, settings)
    i.datadir.mkdir(parents=True)
    (i.datadir / "dirty").touch()
    calls = []
    with pytest.raises(exceptions.CommandError):
        with monkeypatch.context() as m:
            m.setattr("pglift.systemd.enable", lambda *a: calls.append(a))
            postgresql.initdb(ctx, manifest, i)
    assert not i.waldir.exists()
    if ctx.settings.service_manager == "systemd":
        assert not calls


@pytest.mark.parametrize("data_checksums", [True, False])
def test_initdb_force_data_checksums(
    ctx: Context, pg_version: str, data_checksums: bool
) -> None:
    settings = ctx.settings
    assert settings.postgresql.initdb.data_checksums is None
    manifest = interface.Instance(
        name="checksums", version=pg_version, data_checksums=data_checksums
    )
    initdb_options = manifest.initdb_options(settings.postgresql.initdb)
    assert bool(initdb_options.data_checksums) == data_checksums
    instance = system.BaseInstance.get(manifest.name, manifest.version, settings)

    def fake_init(*a: Any, **kw: Any) -> None:
        instance.datadir.mkdir(parents=True)
        (instance.datadir / "postgresql.conf").touch()

    with patch(
        "pgtoolkit.ctl.PGCtl.init", side_effect=fake_init, autospec=True
    ) as init:
        postgresql.initdb(ctx, manifest, instance)
    expected: dict[str, Any] = {
        "waldir": str(instance.waldir),
        "username": "postgres",
        "encoding": "UTF8",
        "auth_local": "peer",
        "auth_host": "password",
        "locale": "C",
    }
    if data_checksums:
        expected["data_checksums"] = True
    pgctl = ctl.pg_ctl(instance.bindir)
    init.assert_called_once_with(pgctl, instance.datadir, **expected)


def test_postgresql_service_name(ctx: Context, instance: system.Instance) -> None:
    assert ctx.hook.postgresql_service_name(ctx=ctx, instance=instance) == "postgresql"


def test_postgresql_editable_conf(ctx: Context, instance: system.Instance) -> None:
    assert ctx.hook.postgresql_editable_conf(ctx=ctx, instance=instance) == "\n".join(
        [
            "port = 999",
            "unix_socket_directories = /socks, /shoes",
            "# backslash_quote = 'safe_encoding'",
        ]
    )


@pytest.mark.usefixtures("nohook")
def test_configuration_precedence(
    pg_version: str,
    ctx: Context,
    instance: system.Instance,
    instance_manifest: interface.Instance,
) -> None:
    """Settings defined in manifest take precedence over postgresql.conf site template."""
    template = "\n".join(
        [
            "bonjour = 'hello, {name}'",
            "max_connections = 101",
            "port=9876",
            "unix_socket_directories = /tmp, /var/run/postgresql",
        ]
    )

    m = interface.Instance(
        name="foo", version=pg_version, settings={"max_connections": 100, "ssl": True}
    )
    configuration = instances.configuration(ctx, m, instance, template=template)
    assert configuration.as_dict() == {
        "bonjour": "hello, foo",
        "lc_messages": "C",
        "lc_monetary": "C",
        "lc_numeric": "C",
        "lc_time": "C",
        "max_connections": 100,
        "port": 9876,
        "ssl": True,
        "unix_socket_directories": "/tmp, /var/run/postgresql",
    }

    m = m._copy_validate({"port": 1234})
    configuration = instances.configuration(ctx, m, instance, template=template)
    assert configuration.as_dict() == {
        "bonjour": "hello, foo",
        "lc_messages": "C",
        "lc_monetary": "C",
        "lc_numeric": "C",
        "lc_time": "C",
        "max_connections": 100,
        "port": 1234,
        "ssl": True,
        "unix_socket_directories": "/tmp, /var/run/postgresql",
    }


@pytest.mark.usefixtures("nohook")
def test_configuration_configure_postgresql(
    ctx: Context, instance: system.Instance, instance_manifest: interface.Instance
) -> None:
    def configuration_changes(m: interface.Instance) -> ConfigChanges:
        configuration = instances.configuration(ctx, m, instance)
        return postgresql.configure_postgresql(
            configuration=configuration, instance=instance
        )

    configdir = instance.datadir
    postgresql_conf = configdir / "postgresql.conf"
    with postgresql_conf.open("w") as f:
        f.write("bonjour_name = 'overridden'\n")

    changes = configuration_changes(
        instance_manifest._copy_validate(
            {
                "settings": dict(
                    instance_manifest.settings,
                    max_connections=100,
                    shared_buffers="10 %",
                    effective_cache_size="5MB",
                ),
                "port": 5433,
            }
        )
    )
    old_shared_buffers, new_shared_buffers = changes.pop("shared_buffers")
    assert old_shared_buffers is None
    assert new_shared_buffers is not None and new_shared_buffers != "10 %"
    assert changes == {
        "bonjour_name": ("overridden", None),
        "cluster_name": (None, "test"),
        "effective_cache_size": (None, "5MB"),
        "lc_messages": (None, "C"),
        "lc_monetary": (None, "C"),
        "lc_numeric": (None, "C"),
        "lc_time": (None, "C"),
        "log_filename": (None, f"{instance.qualname}-%Y-%m-%d_%H%M%S.log"),
        "log_destination": (None, "stderr"),
        "logging_collector": (None, True),
        "max_connections": (None, 100),
        "port": (None, 5433),
        "shared_preload_libraries": (None, "passwordcheck"),
        "unix_socket_directories": (
            None,
            str(ctx.settings.postgresql.socket_directory),
        ),
        "log_directory": (
            None,
            str(ctx.settings.postgresql.logpath),
        ),
    }

    postgresql_conf = configdir / "postgresql.conf"
    content = postgresql_conf.read_text()
    lines = content.splitlines()
    assert "port = 5433" in lines
    assert "cluster_name = 'test'" in lines
    assert re.search(r"shared_buffers = '\d+ [kMGT]?B'", content)
    assert "effective_cache_size" in content
    assert (
        f"unix_socket_directories = '{ctx.settings.prefix}/run/postgresql'" in content
    )

    with postgresql_conf.open() as f:
        config = parse_pgconf(f)
    assert config.port == 5433
    assert config.entries["bonjour_name"].commented
    assert config.cluster_name == "test"

    changes = configuration_changes(
        instance_manifest._copy_validate(
            {
                "settings": dict(
                    instance_manifest.settings,
                    listen_address="*",
                    log_directory="pglogs",
                ),
                "port": 5432,
            }
        )
    )
    old_effective_cache_size, new_effective_cache_size = changes.pop(
        "effective_cache_size"
    )
    assert old_effective_cache_size == "5MB"
    assert new_effective_cache_size != old_effective_cache_size
    old_shared_buffers1, new_shared_buffers1 = changes.pop("shared_buffers")
    assert old_shared_buffers1 == new_shared_buffers
    assert new_shared_buffers1 != old_shared_buffers1
    assert changes == {
        "listen_address": (None, "*"),
        "max_connections": (100, None),
        "port": (5433, None),
        "log_directory": (str(ctx.settings.postgresql.logpath), "pglogs"),
    }

    # Same configuration, no change.
    mtime_before = postgresql_conf.stat().st_mtime
    changes = configuration_changes(
        instance_manifest._copy_validate(
            {
                "settings": dict(
                    instance_manifest.settings,
                    listen_address="*",
                    log_directory="pglogs",
                ),
            }
        )
    )
    assert changes == {}
    mtime_after = postgresql_conf.stat().st_mtime
    assert mtime_before == mtime_after


def test_configure_auth(
    ctx: Context, instance_manifest: interface.Instance, instance: system.Instance
) -> None:
    hba = instance.datadir / "pg_hba.conf"
    ident = instance.datadir / "pg_ident.conf"
    orig_hba = hba.read_text()
    orig_ident = ident.read_text()
    ctx.hook.configure_auth(
        settings=ctx.settings, instance=instance, manifest=instance_manifest
    )
    assert hba.read_text() != orig_hba
    assert ident.read_text() != orig_ident


def test_is_ready(instance: system.Instance) -> None:
    assert not ctl.is_ready(instance)


def test_check_status(instance: system.Instance) -> None:
    with pytest.raises(exceptions.InstanceStateError, match="instance is not_running"):
        postgresql.check_status(instance, Status.running)
    postgresql.check_status(instance, Status.not_running)


def test_start_foreground(ctx: Context, instance: system.Instance) -> None:
    with patch("os.execv", autospec=True) as execv:
        postgresql.start_postgresql(ctx, instance, foreground=True, wait=False)
    postgres = instance.bindir / "postgres"
    execv.assert_called_once_with(
        str(postgres), f"{postgres} -D {instance.datadir}".split()
    )


def test_libpq_environ(settings: Settings, instance: system.Instance) -> None:
    postgres_settings = settings.postgresql
    assert ctl.libpq_environ(instance, postgres_settings.surole.name, base={}) == {
        "PGPASSFILE": str(postgres_settings.auth.passfile)
    }
    assert ctl.libpq_environ(
        instance,
        postgres_settings.surole.name,
        base={"PGPASSFILE": "/var/lib/pgsql/pgpass"},
    ) == {"PGPASSFILE": "/var/lib/pgsql/pgpass"}


def test_libpq_environ_password_command(
    settings: Settings, bindir_template: str, pg_version: str, tmp_path: Path
) -> None:
    s = settings.copy(
        update={
            "postgresql": _postgresql.Settings.parse_obj(
                {
                    "bindir": bindir_template,
                    "surole": {"name": "bob"},
                    "auth": {
                        "password_command": [
                            sys.executable,
                            "-c",
                            "import sys; print(f'{{sys.argv[1]}}-secret')",
                            "{instance}",
                            "--blah",
                        ],
                        "passfile": str(tmp_path / "pgpass"),
                    },
                }
            )
        }
    )
    instance = system.BaseInstance("xyz", pg_version, settings=s)
    assert ctl.libpq_environ(instance, "bob", base={}) == {
        "PGPASSFILE": str(tmp_path / "pgpass"),
        "PGPASSWORD": f"{pg_version}/xyz-secret",
    }


def test_install_systemd_unit_template(
    settings: Settings,
) -> None:
    ((name, content),) = list(
        systemd_unit_templates(
            settings, env={"SETTINGS": "@settings.json"}, content=True
        )
    )
    assert name == "pglift-postgresql@.service"
    lines = content.splitlines()
    assert 'Environment="SETTINGS=@settings.json"' in lines
    for line in lines:
        if line.startswith("ExecStart"):
            execstart = line.split("=", 1)[-1]
            assert execstart == f"{sys.executable} -m pglift postgres %i"
            break
    else:
        raise AssertionError("ExecStart line not found")


def test_logs(
    instance: system.Instance, tmp_path: pathlib.Path, caplog: pytest.LogCaptureFixture
) -> None:
    with pytest.raises(
        exceptions.FileNotFoundError,
        match=r"file 'current_logfiles' for instance \d{2}/test not found",
    ):
        next(ctl.logs(instance))

    current_logfiles = instance.datadir / "current_logfiles"
    current_logfiles.write_text("csvlog log/postgresql.csv\n")
    with pytest.raises(ValueError, match="no record matching 'stderr'"):
        next(ctl.logs(instance, timeout=0.1))

    with (instance.datadir / "postgresql.conf").open("a") as f:
        f.write("\nlog_destination = stderr, csvlog, jsonlog\n")

    stderr_logpath = tmp_path / "postgresql-1.log"
    current_logfiles.write_text(
        f"stderr {stderr_logpath}\n"
        f"csvlog {tmp_path / 'postgresql-1.csv'}\n"
        f"jsonlog {tmp_path / 'postgresql-1.json'}\n"
    )
    with pytest.raises(exceptions.SystemError, match="failed to read"):
        next(ctl.logs(instance, timeout=0.1))

    logger = ctl.logs(instance, timeout=0.1)
    stderr_logpath.write_text("line1\nline2\n")
    caplog.clear()
    with caplog.at_level(logging.INFO, logger="pglift.postgresql.ctl"):
        assert [next(logger) for _ in range(2)] == ["line1\n", "line2\n"]
    assert caplog.messages == [
        f"reading logs of instance {instance} from {stderr_logpath}"
    ]

    with pytest.raises(TimeoutError):
        next(logger)

    logger = ctl.logs(instance)
    assert [next(logger) for _ in range(2)] == ["line1\n", "line2\n"]

    stderr_logpath = tmp_path / "postgresql-2.log"
    current_logfiles.write_text(f"stderr {stderr_logpath}\n")
    stderr_logpath.write_text("line3\nline4\n")

    caplog.clear()
    with caplog.at_level(logging.INFO, logger="pglift.postgresql.ctl"):
        assert [next(logger) for _ in range(2)] == ["line3\n", "line4\n"]
    assert caplog.messages == [
        f"reading logs of instance {instance} from {stderr_logpath}"
    ]


def test_replication_lag(
    instance: system.Instance, standby_instance: system.Instance
) -> None:
    with pytest.raises(TypeError, match="not a standby"):
        postgresql.replication_lag(instance)



from __future__ import annotations

from collections.abc import Sequence
from typing import IO, Any, Callable

import attrs
import click
from pydantic.utils import deep_update

from .. import instances, plugin_manager, postgresql, privileges, task
from ..ctx import Context
from ..models import helpers, interface, system
from ..settings._postgresql import PostgreSQLVersion
from ..types import Status
from .util import (
    Callback,
    CompositeCommandGroup,
    Obj,
    OutputFormat,
    PluggableCommandGroup,
    _list_instances,
    dry_run_option,
    foreground_option,
    instance_lookup,
    output_format_option,
    pass_ctx,
    print_argspec,
    print_json_for,
    print_schema,
    print_table_for,
)


def instance_identifier(
    nargs: int = 1, required: bool = False
) -> Callable[[Callback], Callback]:
    def decorator(fn: Callback) -> Callback:
        command = click.argument(
            "instance",
            nargs=nargs,
            required=required,
            callback=instance_lookup,
            shell_complete=_list_instances,
        )(fn)
        assert command.__doc__
        command.__doc__ += (
            "\n\n    INSTANCE identifies target instance as <version>/<name> where the "
            "<version>/ prefix may be omitted if there is only one instance "
            "matching <name>."
        )
        if not required:
            command.__doc__ += " Required if there is more than one instance on system."
        return command

    return decorator


class InstanceCommands(
    PluggableCommandGroup, CompositeCommandGroup[interface.Instance]
):
    """Group for 'instance' sub-commands handling some of them that require a
    composite interface.Instance model built from registered plugins at
    runtime.
    """

    model = interface.Instance

    def register_plugin_commands(self, obj: Obj) -> None:
        obj.ctx.hook.instance_cli(group=self)


def print_instance_schema(
    context: click.Context, param: click.Parameter, value: bool
) -> None:
    settings = context.obj.ctx.settings
    p = plugin_manager(settings)
    return print_schema(context, param, value, model=interface.Instance.composite(p))


def print_instance_argspec(
    context: click.Context, param: click.Parameter, value: bool
) -> None:
    settings = context.obj.ctx.settings
    model = interface.Instance.composite(plugin_manager(settings))
    print_argspec(context, param, value, model=model)


@click.group(cls=InstanceCommands)
@click.option(
    "--schema",
    is_flag=True,
    callback=print_instance_schema,
    expose_value=False,
    is_eager=True,
    help="Print the JSON schema of instance model and exit.",
)
@click.option(
    "--ansible-argspec",
    is_flag=True,
    callback=print_instance_argspec,
    expose_value=False,
    is_eager=True,
    hidden=True,
    help="Print the Ansible argspec of instance model and exit.",
)
def cli() -> None:
    """Manage instances."""


# Help mypy because click.group() looses the type of 'cls' argument.
assert isinstance(cli, InstanceCommands)


@cli.command_with_composite_model("create")
def _create(Instance: type[interface.Instance]) -> Callback:
    @helpers.parameters_from_model(Instance, "create")
    @click.option(
        "--drop-on-error/--no-drop-on-error",
        default=True,
        help=(
            "On error, drop partially initialized instance by possibly "
            "rolling back operations (true by default)."
        ),
    )
    @pass_ctx
    def command(
        ctx: Context, instance: interface.Instance, drop_on_error: bool
    ) -> None:
        """Initialize a PostgreSQL instance"""
        with ctx.lock:
            if instances.exists(instance.name, instance.version, ctx.settings):
                raise click.ClickException("instance already exists")
            with task.transaction(drop_on_error):
                instances.apply(ctx, instance)

    return command


@cli.command_with_composite_model("alter")
def _alter(Instance: type[interface.Instance]) -> Callback:
    @instance_identifier(nargs=1)
    @helpers.parameters_from_model(Instance, "update", parse_model=False)
    @pass_ctx
    def command(ctx: Context, instance: system.Instance, **changes: Any) -> None:
        """Alter PostgreSQL INSTANCE"""
        with ctx.lock:
            values = instances._get(ctx, instance).dict(exclude={"settings"})
            values = deep_update(values, changes)
            altered = Instance.parse_obj(values)
            instances.apply(ctx, altered)

    return command


@cli.command("apply", hidden=True)
@click.option("-f", "--file", type=click.File("r"), metavar="MANIFEST", required=True)
@output_format_option
@dry_run_option
@pass_ctx
def apply(
    ctx: Context,
    file: IO[str],
    output_format: OutputFormat,
    dry_run: bool,
) -> None:
    """Apply manifest as a PostgreSQL instance"""
    p = plugin_manager(ctx.settings)
    model = interface.Instance.composite(p)
    instance = model.parse_yaml(file)
    if dry_run:
        ret = interface.InstanceApplyResult(change_state=None)
    else:
        with ctx.lock:
            ret = instances.apply(ctx, instance)
    if output_format == OutputFormat.json:
        print_json_for(ret)


@cli.command("promote")
@instance_identifier(nargs=1)
@pass_ctx
def promote(ctx: Context, instance: system.Instance) -> None:
    """Promote standby PostgreSQL INSTANCE"""
    with ctx.lock:
        instances.promote(ctx, instance)


@cli.command("get")
@output_format_option
@instance_identifier(nargs=1)
@pass_ctx
def get(
    ctx: Context,
    instance: system.Instance,
    output_format: OutputFormat,
) -> None:
    """Get the description of PostgreSQL INSTANCE.

    Unless --output-format is specified, 'settings' and 'state' fields are not
    shown as well as 'standby' information if INSTANCE is not a standby.
    """
    exclude = set()
    if not output_format:
        exclude.update(
            [
                "settings",
                "state",
                "data_directory",
                "wal_directory",
                "powa",
            ]
        )
        if not instance.standby:
            exclude.add("standby")
    m = instances.get(ctx, instance).dict(by_alias=True, exclude=exclude)
    if output_format == OutputFormat.json:
        print_json_for(m)
    else:
        print_table_for([m], box=None)


@cli.command("list")
@click.option(
    "--version",
    type=click.Choice(list(PostgreSQLVersion)),
    help="Only list instances of specified version.",
)
@output_format_option
@pass_ctx
def ls(
    ctx: Context,
    version: PostgreSQLVersion | None,
    output_format: OutputFormat,
) -> None:
    """List the available instances"""

    values = [attrs.asdict(i) for i in instances.ls(ctx.settings, version=version)]
    if output_format == OutputFormat.json:
        print_json_for(values)
    else:
        print_table_for(values)


@cli.command("drop")
@instance_identifier(nargs=-1)
@pass_ctx
def drop(ctx: Context, instance: tuple[system.Instance, ...]) -> None:
    """Drop PostgreSQL INSTANCE"""
    with ctx.lock:
        for i in instance:
            instances.drop(ctx, i)


@cli.command("status")
@instance_identifier(nargs=1)
@click.pass_context
def status(context: click.Context, instance: system.Instance) -> None:
    """Check the status of instance and all satellite components.

    Output the status string value ('running', 'not running') for each component.
    If not all services are running, the command exit code will be 3.
    """
    ctx: Context = context.obj.ctx
    exit_code = Status.running.value
    for status, component in reversed(
        ctx.hook.instance_status(ctx=ctx, instance=instance)
    ):
        if status != Status.running:
            exit_code = Status.not_running.value
        click.echo(f"{component}: {status.name.replace('_', ' ')}")
    context.exit(exit_code)


@cli.command("start")
@instance_identifier(nargs=-1)
@foreground_option
@click.option("--all", "all_instances", is_flag=True, help="Start all instances.")
@pass_ctx
def start(
    ctx: Context,
    instance: tuple[system.Instance, ...],
    foreground: bool,
    all_instances: bool,
) -> None:
    """Start PostgreSQL INSTANCE"""
    if foreground and len(instance) != 1:
        raise click.UsageError(
            "only one INSTANCE argument may be given with --foreground"
        )
    with ctx.lock:
        for i in instance:
            instances.start(ctx, i, foreground=foreground)


@cli.command("stop")
@instance_identifier(nargs=-1)
@click.option("--all", "all_instances", is_flag=True, help="Stop all instances.")
@pass_ctx
def stop(
    ctx: Context, instance: tuple[system.Instance, ...], all_instances: bool
) -> None:
    """Stop PostgreSQL INSTANCE"""
    with ctx.lock:
        for i in instance:
            instances.stop(ctx, i)


@cli.command("reload")
@instance_identifier(nargs=-1)
@click.option("--all", "all_instances", is_flag=True, help="Reload all instances.")
@pass_ctx
def reload(
    ctx: Context, instance: tuple[system.Instance, ...], all_instances: bool
) -> None:
    """Reload PostgreSQL INSTANCE"""
    with ctx.lock:
        for i in instance:
            instances.reload(ctx, i)


@cli.command("restart")
@instance_identifier(nargs=-1)
@click.option("--all", "all_instances", is_flag=True, help="Restart all instances.")
@pass_ctx
def restart(
    ctx: Context, instance: tuple[system.Instance, ...], all_instances: bool
) -> None:
    """Restart PostgreSQL INSTANCE"""
    with ctx.lock:
        for i in instance:
            instances.restart(ctx, i)


@cli.command("exec")
@instance_identifier(nargs=1, required=True)
@click.argument("command", required=True, nargs=-1, type=click.UNPROCESSED)
@pass_ctx
def exec(ctx: Context, instance: system.Instance, command: tuple[str, ...]) -> None:
    """Execute command in the libpq environment for PostgreSQL INSTANCE.

    COMMAND parts may need to be prefixed with -- to separate them from
    options when confusion arises.
    """
    instances.exec(ctx, instance, command)


@cli.command("env")
@instance_identifier(nargs=1)
@output_format_option
@pass_ctx
def env(ctx: Context, instance: system.Instance, output_format: OutputFormat) -> None:
    """Output environment variables suitable to handle to PostgreSQL INSTANCE.

    This can be injected in shell using:

        export $(pglift instance env myinstance)
    """
    instance_env = instances.env_for(ctx, instance, path=True)
    if output_format == OutputFormat.json:
        print_json_for(instance_env)
    else:
        for key, value in sorted(instance_env.items()):
            click.echo(f"{key}={value}")


@cli.command("logs")
@click.option("--follow/--no-follow", "-f/", default=False, help="Follow log output.")
@instance_identifier(nargs=1)
def logs(instance: system.Instance, follow: bool) -> None:
    """Output PostgreSQL logs of INSTANCE.

    This assumes that the PostgreSQL instance is configured to use file-based
    logging (i.e. log_destination amongst 'stderr' or 'csvlog').
    """
    if follow:
        logstream = postgresql.logs(instance)
    else:
        logstream = postgresql.logs(instance, timeout=0)
    try:
        for line in logstream:
            click.echo(line, nl=False)
    except TimeoutError:
        pass
    except FileNotFoundError as e:
        raise click.ClickException(str(e)) from None


@cli.command("privileges")
@instance_identifier(nargs=1)
@click.option(
    "-d",
    "--database",
    "databases",
    multiple=True,
    help="Database to inspect. When not provided, all databases are inspected.",
)
@click.option("-r", "--role", "roles", multiple=True, help="Role to inspect")
@click.option("--default", "defaults", is_flag=True, help="Display default privileges")
@output_format_option
@pass_ctx
def list_privileges(
    ctx: Context,
    instance: system.Instance,
    databases: Sequence[str],
    roles: Sequence[str],
    defaults: bool,
    output_format: OutputFormat,
) -> None:
    """List privileges on INSTANCE's databases."""
    with postgresql.running(ctx, instance):
        try:
            prvlgs = privileges.get(
                ctx, instance, databases=databases, roles=roles, defaults=defaults
            )
        except ValueError as e:
            raise click.ClickException(str(e)) from None
    values = [attrs.asdict(p) for p in prvlgs]
    if output_format == OutputFormat.json:
        print_json_for(values)
    else:
        if defaults:
            title = f"Default privileges on instance {instance}"
        else:
            title = f"Privileges on instance {instance}"
        print_table_for(values, title=title)


@cli.command("upgrade")
@instance_identifier(nargs=1)
@click.option(
    "--version",
    "newversion",
    type=click.Choice(list(PostgreSQLVersion)),
    help="PostgreSQL version of the new instance (default to site-configured value).",
)
@click.option(
    "--name", "newname", help="Name of the new instance (default to old instance name)."
)
@click.option(
    "--port", required=False, type=click.INT, help="Port of the new instance."
)
@click.option(
    "--jobs",
    required=False,
    type=click.INT,
    help="Number of simultaneous processes or threads to use (from pg_upgrade).",
)
@pass_ctx
def upgrade(
    ctx: Context,
    instance: system.Instance,
    newversion: PostgreSQLVersion | None,
    newname: str | None,
    port: int | None,
    jobs: int | None,
) -> None:
    """Upgrade INSTANCE using pg_upgrade"""
    with ctx.lock:
        postgresql.check_status(instance, Status.not_running)
        new_instance = instances.upgrade(
            ctx, instance, version=newversion, name=newname, port=port, jobs=jobs
        )
        instances.start(ctx, new_instance)



from __future__ import annotations

import socket
import subprocess
from dataclasses import dataclass
from pathlib import Path

from tenacity import retry
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_fixed
from typing_extensions import Self


@dataclass
class SSHKeys:
    host_key: Path
    private_key: Path
    public_key: Path

    @classmethod
    def make(cls, path: Path) -> Self:
        (path / "etc" / "ssh").mkdir(parents=True, exist_ok=True)
        subprocess.run(
            ["ssh-keygen", "-A", "-f", path], check=True, capture_output=True
        )
        host_key = path / "etc" / "ssh" / "ssh_host_rsa_key"

        private_key = path / "id_rsa"
        subprocess.run(
            [
                "ssh-keygen",
                "-f",
                str(private_key),
                "-t",
                "rsa",
                "-b",
                "2048",
                "-N",
                "",
            ],
            check=True,
            capture_output=True,
        )
        public_key = path / "id_rsa.pub"

        return cls(host_key, private_key, public_key)


def add_to_known_hosts(path: Path, hostname: str, port: int) -> None:
    @retry(
        retry=retry_if_exception_type(ConnectionRefusedError),
        wait=wait_fixed(1),
        stop=stop_after_attempt(5),
    )
    def try_connect(hostname: str, port: int) -> None:
        with socket.socket() as s:
            s.connect((hostname, port))

    with open(path / "known_hosts", "w") as f:
        try_connect(hostname, port)
        subprocess.run(["ssh-keyscan", "-p", str(port), hostname], stdout=f, check=True)



from __future__ import annotations

import pathlib
from collections.abc import Iterator
from pathlib import Path
from typing import Any, Callable
from unittest.mock import MagicMock, patch

import pydantic
import pytest

from pglift import _install, plugin_manager
from pglift.ctx import Context
from pglift.models import interface
from pglift.models.system import Instance
from pglift.pgbackrest import models as pgbackrest_models
from pglift.pm import PluginManager
from pglift.prometheus import impl as prometheus_mod
from pglift.prometheus import models as prometheus_models
from pglift.settings import Settings, _postgresql, _systemd
from pglift.temboard import impl as temboard_mod
from pglift.temboard import models as temboard_models


def pytest_addoption(parser: Any) -> None:
    parser.addoption(
        "--write-changes",
        action="store_true",
        default=False,
        help="Write-back changes to test data.",
    )


@pytest.fixture
def write_changes(request: Any) -> bool:
    value = request.config.option.write_changes
    assert isinstance(value, bool)
    return value


@pytest.fixture
def pg_version() -> str:
    return "15"


@pytest.fixture
def pgbackrest_execpath(tmp_path: Path) -> Path:
    execpath = tmp_path / "pgbackrest"
    execpath.touch(0o700)
    execpath.write_text("#!/bin/sh\nexit 1\n")
    return execpath


@pytest.fixture
def prometheus_execpath(tmp_path: Path) -> Path:
    execpath = tmp_path / "postgres_exporter"
    execpath.touch(0o700)
    execpath.write_text("#!/bin/sh\nexit 1\n")
    return execpath


@pytest.fixture
def temboard_execpath(tmp_path: Path) -> Path:
    execpath = tmp_path / "temboard-agent"
    execpath.touch(0o700)
    execpath.write_text("#!/bin/sh\nexit 1\n")
    return execpath


@pytest.fixture
def patroni_execpath(tmp_path: Path) -> Path:
    execpath = tmp_path / "patroni"
    execpath.touch(0o700)
    execpath.write_text(
        "\n".join(
            [
                "#!/bin/sh",
                'if [ "$1" = "--validate-config" ]',
                "then",
                "  echo 'test test test'",
                "fi",
                "exit 1",
            ]
        )
    )
    return execpath


@pytest.fixture
def systemctl(tmp_path: Path) -> Iterator[str]:
    p = str(tmp_path / "systemctl")
    with patch("pglift.settings._systemd.Settings.systemctl", create=True, new=p):
        yield p


@pytest.fixture
def systemd_settings(tmp_path: Path, systemctl: str) -> _systemd.Settings:
    return _systemd.Settings.parse_obj({"unit_path": str(tmp_path / "systemd")})


@pytest.fixture
def settings(
    tmp_path: Path,
    patroni_execpath: Path,
    pgbackrest_execpath: Path,
    prometheus_execpath: Path,
    temboard_execpath: Path,
    systemd_settings: _systemd.Settings,
) -> Settings:
    passfile = tmp_path / "pgass"
    passfile.touch()
    signing_key = tmp_path / "signing-public.pem"
    signing_key.touch()
    ssl_cert_file = tmp_path / "temboard-agent.pem"
    ssl_cert_file.touch()
    ssl_key_file = tmp_path / "temboard-agent.key"
    ssl_key_file.touch()
    ssl_ca_cert_file = tmp_path / "mycacert.pem"
    ssl_ca_cert_file.touch()

    def pg_bindir(path: Path, version: int) -> Path:
        """Make a bindir with minimal versions of pg_* executables.

        - pg_ctl should work for --version and status
        - pg_checksums returns 0, as it's invoked upon instances._get()
        - psql exists, but always fails
        - pg_isready returns 1
        """
        path.mkdir(parents=True)
        pg_ctl = path / "pg_ctl"
        pg_ctl.touch(mode=0o700)
        pg_ctl.write_text(
            "\n".join(
                [
                    "#!/bin/sh",
                    "case $1 in",
                    "  --version)",
                    f"     echo 'pg_ctl (PostgreSQL) {version}.1'",
                    "     exit 0",
                    "     ;;",
                    "  status)",
                    "     exit 3",  # not running
                    "     ;;",
                    "  *)",
                    '     echo "unexpected invocation pg_ctl $1" >&2',
                    "     exit 1",
                    "     ;;",
                    "esac",
                ]
            )
        )

        pg_checksums = path / "pg_checksums"
        pg_checksums.touch(mode=0o700)
        pg_checksums.write_text("#!/bin/sh\nexit 0")

        pg_controldata = path / "pg_controldata"
        pg_controldata.touch(mode=0o700)
        pg_controldata.write_text(
            "\n".join(
                [
                    "#!/bin/sh",
                    'echo "Data page checksum version:           1"',
                    "exit 0",
                ]
            )
        )

        psql = path / "psql"
        psql.touch(mode=0o700)
        psql.write_text(
            "\n".join(
                ["#!/bin/sh", 'echo "unexpected psql $* invocation" >&2', "exit 1"]
            )
        )

        pg_isready = path / "pg_isready"
        pg_isready.touch(mode=0o700)
        pg_isready.write_text("#!/bin/sh\nexit 1")

        return path

    pg13_bindir = pg_bindir(tmp_path / "pgsql-13" / "bin", 13)
    pg_bindir(tmp_path / "postgresql" / "15" / "bin", 15)
    base_bindir = str(tmp_path / "postgresql" / "{version}" / "bin")
    patroni_restapi_certfile = tmp_path / "patroni-restapi.pem"
    patroni_restapi_certfile.touch()
    crldir = tmp_path / "crls"
    crldir.mkdir()
    obj = {
        "prefix": str(tmp_path),
        "run_prefix": str(tmp_path / "run"),
        "postgresql": {
            "auth": {
                "local": "peer",
                "host": "password",
                "passfile": str(passfile),
            },
            "bindir": base_bindir,
            "default_version": "15",
            "versions": [
                {"version": "13", "bindir": str(pg13_bindir)},
            ],
            "replrole": "replication",
        },
        "systemd": systemd_settings,
        "service_manager": None,
        "scheduler": None,
        "patroni": {
            "execpath": patroni_execpath,
            "etcd": {
                "hosts": ["etcd1:123", "etcd2:456"],
            },
            "postgresql": {
                "connection": {
                    "ssl": {
                        "mode": "verify-ca",
                        "crldir": crldir,
                    },
                },
                "use_pg_rewind": True,
            },
            "restapi": {
                "certfile": patroni_restapi_certfile,
                "verify_client": "optional",
            },
        },
        "pgbackrest": {
            "execpath": pgbackrest_execpath,
            "repository": {"mode": "path", "path": tmp_path / "backups"},
        },
        "prometheus": {"execpath": prometheus_execpath},
        "logrotate": {},
        "rsyslog": {},
        "temboard": {
            "execpath": temboard_execpath,
            "ui_url": "https://0.0.0.0:8888",
            "signing_key": str(signing_key),
            "certificate": {
                "ca_cert": str(ssl_ca_cert_file),
                "cert": str(ssl_cert_file),
                "key": str(ssl_key_file),
            },
        },
        "powa": {},
        "cli": {
            "log_format": "%(levelname)-4s %(message)s",
        },
    }
    settings = Settings.parse_obj(obj)
    assert {v.version for v in settings.postgresql.versions} == {"13", "15"}
    return settings


@pytest.fixture
def ctx(settings: Settings) -> Context:
    return Context(settings=settings)


@pytest.fixture
def pm(settings: Settings) -> PluginManager:
    return plugin_manager(settings)


@pytest.fixture
def nohook(pm: PluginManager) -> Iterator[None]:
    unregistered = pm.unregister_all()
    yield
    for plugin in unregistered:
        pm.register(plugin)


@pytest.fixture
def installed(
    settings: Settings, pm: PluginManager, ctx: Context, systemctl: str
) -> Iterator[None]:
    _install.do(pm, settings, header="pglift's unit tests")
    yield
    _install.undo(pm, settings, ctx)


@pytest.fixture
def composite_instance_model(pm: PluginManager) -> type[interface.Instance]:
    return interface.Instance.composite(pm)


@pytest.fixture
def composite_role_model(pm: PluginManager) -> type[interface.Role]:
    return interface.Role.composite(pm)


@pytest.fixture
def instance_manifest(
    composite_instance_model: type[interface.Instance], pg_version: str
) -> interface.Instance:
    return composite_instance_model(
        name="test",
        version=pg_version,
        surole_password=pydantic.SecretStr("p0st.g're$"),
        replrole_password=pydantic.SecretStr("repl1&c"),
        settings={"shared_preload_libraries": "passwordcheck"},
        pgbackrest={"stanza": "test-stanza"},
    )


def _instance(
    name: str, version: str, postgresql_conf: str, settings: Settings
) -> Instance:
    # Services are looked-up in reverse order of plugin registration.
    services: list[Any] = []

    assert settings.temboard is not None
    temboard_port = 2345
    temboard = temboard_models.Service(
        name=f"{version}-{name}",
        settings=settings.temboard,
        port=temboard_port,
        password=pydantic.SecretStr("dorade"),
    )
    services.append(temboard)

    assert settings.prometheus is not None
    prometheus_port = 9817
    prometheus = prometheus_models.Service(
        name=f"{version}-{name}",
        settings=settings.prometheus,
        port=prometheus_port,
        password=pydantic.SecretStr("truite"),
    )
    services.append(prometheus)

    assert settings.pgbackrest is not None
    pgbackrest = pgbackrest_models.Service(
        stanza=f"{name}-stanza",
        path=settings.pgbackrest.configpath / "conf.d" / f"{name}-stanza.conf",
    )
    services.append(pgbackrest)

    instance = Instance(
        name=name,
        version=version,
        settings=settings,
        services=services,
    )
    assert not instance.datadir.exists()
    instance.datadir.mkdir(parents=True)
    (instance.datadir / "PG_VERSION").write_text(instance.version)
    (instance.datadir / "postgresql.conf").write_text(postgresql_conf)
    (instance.datadir / "pg_hba.conf").write_text(
        "# pg_hba.conf\nlocal all postgres peer\n"
    )
    (instance.datadir / "pg_ident.conf").write_text("# pg_ident.conf\nmymap test dba\n")

    prometheus_config = prometheus_mod._configpath(
        instance.qualname, settings.prometheus
    )
    assert not prometheus_config.exists()
    prometheus_config.parent.mkdir(parents=True, exist_ok=True)
    prometheus_config.write_text(
        f"DATA_SOURCE_NAME=dbname=postgres port={instance.port} host={settings.postgresql.socket_directory} user=monitoring sslmode=disable password=truite\n"
        f"PG_EXPORTER_WEB_LISTEN_ADDRESS=:{prometheus.port}"
    )

    temboard_config = temboard_mod._configpath(instance.qualname, settings.temboard)
    assert not temboard_config.exists()
    temboard_config.parent.mkdir(parents=True, exist_ok=True)
    temboard_config.write_text(
        "\n".join(
            [
                "[temboard]",
                f"port = {temboard.port}",
                "ui_url = https://0.0.0.0:8888",
                "[postgresql]",
                f"port = {instance.port}",
                f"host = {settings.postgresql.socket_directory}",
                "user = temboardagent",
                "password = dorade",
            ]
        )
    )

    stanza_config = pgbackrest.path
    assert not stanza_config.exists()
    stanza_config.parent.mkdir(parents=True, exist_ok=True)
    stanza_config.write_text(
        "\n".join(
            [
                f"[{pgbackrest.stanza}]",
                f"pg1-path = {instance.datadir}",
                f"pg1-port = {instance.port}",
                "pg1-user = backup",
            ]
        )
    )

    assert settings.logrotate
    settings.logrotate.configdir.mkdir(parents=True, exist_ok=True)
    assert settings.rsyslog
    settings.rsyslog.configdir.mkdir(parents=True, exist_ok=True)
    return instance


@pytest.fixture
def postgresql_conf() -> str:
    return "\n".join(
        [
            "port = 999",
            "unix_socket_directories = /socks, /shoes",
            "# backslash_quote = 'safe_encoding'",
        ]
    )


@pytest.fixture
def instance(pg_version: str, postgresql_conf: str, settings: Settings) -> Instance:
    return _instance("test", pg_version, postgresql_conf, settings)


@pytest.fixture
def instance2(pg_version: str, postgresql_conf: str, settings: Settings) -> Instance:
    return _instance("test2", pg_version, postgresql_conf, settings)


@pytest.fixture
def standby_instance(
    pg_version: str, postgresql_conf: str, settings: Settings
) -> Instance:
    instance = _instance("standby", pg_version, postgresql_conf, settings)
    assert int(pg_version) >= 12
    (instance.datadir / "standby.signal").write_text("")
    (instance.datadir / "postgresql.auto.conf").write_text(
        "primary_conninfo = 'host=/tmp port=4242 user=pg'\n"
        "primary_slot_name = aslot\n"
    )
    return instance


@pytest.fixture
def meminfo(tmp_path: Path) -> Path:
    fpath = tmp_path / "meminfo"
    fpath.write_text(
        "\n".join(
            [
                "MemTotal:        6022056 kB",
                "MemFree:         3226640 kB",
                "MemAvailable:    4235060 kB",
                "Buffers:          206512 kB",
            ]
        )
    )
    return fpath


@pytest.fixture(scope="package")
def site_config(
    site_config: Callable[..., str | None]
) -> Iterator[Callable[..., str | None]]:
    def test_site_config(*args: str) -> str | None:
        """Lookup for configuration files in local data director first."""
        datadir = pathlib.Path(__file__).parent / "data"
        fpath = datadir.joinpath(*args)
        if fpath.exists():
            return fpath.read_text()
        return site_config(*args)

    with patch("pglift.util.site_config", new=test_site_config) as fn:
        yield fn


@pytest.fixture(scope="function", autouse=True)
def run_as_non_root() -> Iterator[MagicMock]:
    with patch("pglift.settings.is_root", autospec=True, return_value=False) as m:
        yield m


@pytest.fixture(scope="function")
def run_as_root(run_as_non_root: MagicMock) -> None:
    run_as_non_root.return_value = True


@pytest.fixture
def bindir_template(tmp_path: Path) -> str:
    bindir = tmp_path / "15" / "bin"
    bindir.mkdir(parents=True)
    return str(tmp_path / "{version}" / "bin")


@pytest.fixture
def postgresql_settings(bindir_template: str) -> _postgresql.Settings:
    return _postgresql.Settings(bindir=bindir_template)


@pytest.fixture
def default_settings(
    tmp_path: Path, postgresql_settings: _postgresql.Settings
) -> Settings:
    return Settings(prefix=tmp_path, postgresql=postgresql_settings)



from __future__ import annotations

from typing import Literal

from .. import hookimpl
from ..ctx import Context
from ..settings import Settings
from . import disable, enable, start, stop


def register_if(settings: Settings) -> bool:
    return settings.scheduler == "systemd"


def unit(service: str, qualname: str) -> str:
    return f"pglift-{service}@{qualname}.timer"


@hookimpl
def schedule_service(ctx: Context, service: str, name: str) -> Literal[True]:
    enable(ctx, unit(service, name))
    return True


@hookimpl
def unschedule_service(
    ctx: Context, service: str, name: str, now: bool | None
) -> Literal[True]:
    kwargs = {}
    if now is not None:
        kwargs["now"] = now
    disable(ctx, unit(service, name), **kwargs)
    return True


@hookimpl
def start_timer(ctx: Context, service: str, name: str) -> Literal[True]:
    start(ctx, unit(service, name))
    return True


@hookimpl
def stop_timer(ctx: Context, service: str, name: str) -> Literal[True]:
    stop(ctx, unit(service, name))
    return True



from __future__ import annotations

import logging
import socket
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pydantic
import pytest
import yaml
from click.testing import CliRunner

from pglift import exceptions, instances
from pglift.cli.util import CLIContext, Obj
from pglift.ctx import Context
from pglift.models import interface, system
from pglift.patroni import (
    configure_postgresql,
    impl,
    instance_env,
    models,
    systemd_unit_templates,
)
from pglift.patroni.cli import cli as cli
from pglift.settings import Settings, _patroni

from . import click_result_traceback


@pytest.fixture
def patroni_settings(settings: Settings) -> _patroni.Settings:
    assert settings.patroni
    return settings.patroni


@pytest.fixture
def instance_manifest(
    composite_instance_model: type[interface.Instance],
) -> interface.Instance:
    return composite_instance_model.parse_obj(
        {
            "name": "test",
            "version": "12",
            "data_checksums": True,
            "settings": {
                "shared_buffers": "257MB",
                "effective_cache_size": "4 GB",
                "unix_socket_directories": "/tmp/tests",
                "log_connections": "on",
                "log_directory": "/tmp/log",
                "log_filename": "patgres-%Y-%m-%d.log",
                "log_disconnections": "false",
                "log_checkpoints": True,
                "log_min_duration_statement": "3s",
                "shared_preload_libraries": "passwordcheck",
            },
            "surole_password": None,
            "replrole_password": "rrr",
            "patroni": {"cluster": "whatever"},
            "pgbackrest": {"stanza": "test-stanza"},
        }
    )


@pytest.fixture
def instance(
    datadir: Path, patroni_settings: _patroni.Settings, instance: system.Instance
) -> system.Instance:
    patroni_config = impl._configpath(instance.qualname, patroni_settings)
    patroni_config.parent.mkdir(parents=True, exist_ok=True)
    patroni_config.write_text((datadir / "patroni.yaml").read_text())
    impl._pgpass(instance.qualname, patroni_settings.postgresql).write_text("# test\n")
    patroni = models.Service(
        cluster="test-scope",
        node="pg1",
        name=instance.qualname,
        settings=patroni_settings,
    )
    instance.services.append(patroni)
    return instance


def test_servicemanifest_defaults() -> None:
    s = models.ServiceManifest(cluster="cluster", node=None, restapi=None)
    assert s.node == socket.getfqdn()
    assert set(s.restapi.dict()) == {"connect_address", "listen"}


def test_available(settings: Settings) -> None:
    assert impl.available(settings)


def test_systemd_unit_templates(settings: Settings, patroni_execpath: Path) -> None:
    ((name, content),) = list(systemd_unit_templates(settings=settings, content=True))
    assert name == "pglift-patroni@.service"
    assert (
        f"ExecStart={patroni_execpath} {settings.prefix}/etc/patroni/%i.yaml" in content
    )


def test_patroni_incompatible_with_standby(
    composite_instance_model: type[interface.Instance],
) -> None:
    with pytest.raises(pydantic.ValidationError):
        composite_instance_model.parse_obj(
            {
                "name": "invalid",
                "standby": {"primary_conninfo": "port=5444"},
                "patroni": {"cluster": "tests"},
            }
        )


def test_patroni_extra_allow() -> None:
    p = models.Patroni.parse_obj(
        {
            "scope": "foo",
            "name": "bar",
            "postgresql": {
                "connect_address": "db.host:4378",
                "listen": "122.78.9.6:5433",
                "parameters": {"log_connections": True},
                "callbacks": {"on_reload": "/my/scripts/reload.sh"},
            },
            "restapi": {
                "connect_address": "api.server:8088",
                "listen": "122.76.9.1:8808",
                "http_extra_headers": {"X-Frame-Options": "SAMEORIGIN"},
            },
            "tags": {"nosync": True},
        }
    ).dict()
    assert p == {
        "scope": "foo",
        "name": "bar",
        "restapi": {
            "connect_address": "api.server:8088",
            "listen": "122.76.9.1:8808",
            "cafile": None,
            "certfile": None,
            "keyfile": None,
            "verify_client": None,
            "http_extra_headers": {"X-Frame-Options": "SAMEORIGIN"},
        },
        "postgresql": {
            "connect_address": "db.host:4378",
            "listen": "122.78.9.6:5433",
            "parameters": {"log_connections": True},
            "callbacks": {"on_reload": "/my/scripts/reload.sh"},
        },
        "tags": {"nosync": True},
    }


def test_servicemanifest_extra_forbid() -> None:
    with pytest.raises(pydantic.ValidationError) as excinfo:
        models.ServiceManifest.parse_obj(
            {
                "cluster": "foo",
                "node": "bar",
                "postgresql": {
                    "connect_host": "db.host:4378",
                    "listen": "122.78.9.6:5433",
                    "parameters": {"log_connections": True},
                },
                "restapi": {
                    "connect_address": "api.server:8088",
                    "listen": "122.76.9.1:8808",
                    "http_extra_headers": {"X-Frame-Options": "SAMEORIGIN"},
                },
                "tags": {"nosync": True},
            }
        )
    assert excinfo.value.errors() == [
        {
            "loc": ("restapi", "http_extra_headers"),
            "msg": "extra fields not permitted",
            "type": "value_error.extra",
        },
        {
            "loc": ("postgresql", "listen"),
            "msg": "extra fields not permitted",
            "type": "value_error.extra",
        },
        {
            "loc": ("postgresql", "parameters"),
            "msg": "extra fields not permitted",
            "type": "value_error.extra",
        },
        {
            "loc": ("tags",),
            "msg": "extra fields not permitted",
            "type": "value_error.extra",
        },
    ]


@pytest.fixture
def patroni(
    ctx: Context,
    tmp_path: Path,
    patroni_settings: _patroni.Settings,
    instance: system.Instance,
    instance_manifest: interface.Instance,
) -> models.Patroni:
    m = instance_manifest._copy_validate(
        {
            "data_checksums": True,
            "surole_password": None,
            "replrole_password": "rrr",
        }
    )
    configuration = instances.configuration(ctx, instance_manifest, instance)
    cacert = tmp_path / "cacert.pem"
    cacert.touch()
    cert = tmp_path / "host.pem"
    cert.touch()
    key = tmp_path / "host.key"
    key.touch()

    with patch(
        "socket.gethostbyname",
        side_effect=AssertionError("gethostbyname unexpectedly called"),
        autospec=True,
    ), patch(
        "pglift.pgbackrest.base.backup_info", return_value={"backup": []}, autospec=True
    ) as backup_info:
        svc = models.ServiceManifest.parse_obj(
            {
                "cluster": "whatever, will be overridden",
                "restapi": {"connect_address": "api.server:8999"},
                "postgresql": {
                    "connect_host": "pghost.test",
                    "replication": {
                        "ssl": {"cert": cert, "key": key, "password": "repsslpwd"}
                    },
                },
            }
        )
        p = models.Patroni.build(
            patroni_settings,
            svc,
            instance,
            m,
            configuration,
            scope="test-scope",
            name="pg1",
            etcd3=(
                patroni_settings.etcd.dict(exclude={"v2"})
                | {
                    "protocol": "https",
                    "cacert": "/path/to/cacert.pem",
                    "cert": "/path/to/host.pem",
                    "key": "/path/to/host.key",
                }
            ),
            watchdog={
                "mode": "required",
                "device": "/dev/watchdog",
                "safety_margin": 5,
            },
            restapi={
                "connect_address": "localhost:8080",
                "cafile": cacert,
                "certfile": cert,
                "keyfile": key,
                "verify_client": "optional",
            },
            ctl={
                "certfile": "/path/to/host.pem",
                "keyfile": "/path/to/host.key",
            },
            postgresql={"pgpass": "/path/to/patroni/test.pgpass"},
        )
    assert backup_info.call_count == 1
    return p


def test_yaml(
    patroni: models.Patroni, datadir: Path, write_changes: bool, tmp_path: Path
) -> None:
    def fake_path_encoder(obj: Any) -> Any:
        if isinstance(obj, Path):
            return obj.name
        raise TypeError

    doc = patroni.yaml(encoder=fake_path_encoder)
    fpath = datadir / "patroni.yaml"
    if write_changes:
        fpath.write_text(doc)

    expected = fpath.read_text()
    assert doc == expected


def test_validate_config(
    patroni: models.Patroni,
    patroni_settings: _patroni.Settings,
    caplog: pytest.LogCaptureFixture,
) -> None:
    with caplog.at_level(logging.WARNING):
        impl.validate_config(patroni.yaml(), patroni_settings)
    (msg,) = caplog.messages
    assert msg.strip() == "invalid Patroni configuration: test test test"


def test_maybe_backup_config(
    instance: system.Instance,
    patroni_settings: _patroni.Settings,
    caplog: pytest.LogCaptureFixture,
) -> None:
    with patch.object(
        impl,
        "cluster_members",
        return_value=[
            models.ClusterMember(
                host="h", name="node", port=8097, role="leader", state="s"
            )
        ],
        autospec=True,
    ):
        with caplog.at_level("WARNING", logger="pglift.patroni"):
            impl.maybe_backup_config(
                instance.qualname, node="node", cluster="clu", settings=patroni_settings
            )
    assert (
        "'node' appears to be the last member of cluster 'clu', saving Patroni configuration file"
        in caplog.messages[0]
    )
    backuppath = next(patroni_settings.configpath.parent.glob("clu-node*.yaml"))
    backupconfig = yaml.safe_load(backuppath.read_text())
    assert backupconfig["etcd3"] == {
        "cacert": "/path/to/cacert.pem",
        "cert": "/path/to/host.pem",
        "hosts": ["etcd1:123", "etcd2:456"],
        "key": "/path/to/host.key",
        "protocol": "https",
    }
    assert backupconfig["restapi"] == {
        "connect_address": "localhost:8080",
        "cafile": "cacert.pem",
        "certfile": "host.pem",
        "keyfile": "host.key",
        "listen": "localhost:8080",
        "verify_client": "optional",
    }
    assert backupconfig["ctl"] == {
        "certfile": "/path/to/host.pem",
        "keyfile": "/path/to/host.key",
    }
    pgpass = next(patroni_settings.configpath.parent.glob("clu-node*.pgpass"))
    assert pgpass.read_text() == "# test\n"


def test_postgresql_service_name(ctx: Context, instance: system.Instance) -> None:
    assert ctx.hook.postgresql_service_name(ctx=ctx, instance=instance) == "patroni"


def test_postgresql_editable_conf(ctx: Context, instance: system.Instance) -> None:
    assert (
        ctx.hook.postgresql_editable_conf(ctx=ctx, instance=instance)
        == "\n".join(
            [
                "archive_command = 'pgbackrest --config-path=/cfg/pgbackrest --stanza=test-stanza --pg1-path=/pg/data archive-push %p'",
                "archive_mode = on",
                "cluster_name = 'test'",
                "effective_cache_size = '4 GB'",
                "lc_messages = 'C'",
                "lc_monetary = 'C'",
                "lc_numeric = 'C'",
                "lc_time = 'C'",
                "log_checkpoints = on",
                "log_connections = on",
                "log_destination = 'syslog'",
                "log_directory = '/tmp/log'",
                "log_disconnections = off",
                "log_filename = 'patgres-%Y-%m-%d.log'",
                "log_min_duration_statement = '3s'",
                "logging_collector = on",
                "shared_buffers = '257MB'",
                "shared_preload_libraries = 'passwordcheck, pg_qualstats, pg_stat_statements, pg_stat_kcache'",
                "syslog_ident = 'postgresql-15-test'",
                "unix_socket_directories = '/tmp/tests'",
                "wal_level = 'replica'",
            ]
        )
        + "\n"
    )


def test_configure_postgresql(
    ctx: Context,
    patroni_settings: _patroni.Settings,
    instance_manifest: interface.Instance,
    instance: system.Instance,
) -> None:
    instance_manifest.settings["work_mem"] = "8MB"
    pgconfig = instances.configuration(ctx, instance_manifest, instance)
    patroni = models.Patroni.get(instance.qualname, patroni_settings)
    postgresql_parameters = patroni.postgresql.parameters.copy()
    assert "work_mem" not in postgresql_parameters
    postgresql_parameters["work_mem"] = "8MB"
    with patch.object(impl, "api_request", autospec=True) as api_request:
        changes = configure_postgresql(ctx, instance_manifest, pgconfig, instance)
    assert changes == {"work_mem": (None, "8MB")}
    (p, verb, endpoint), _ = api_request.call_args
    assert (verb, endpoint) == ("POST", "reload")
    assert p.postgresql.parameters == postgresql_parameters


def test_env(settings: Settings, instance: system.Instance, pg_version: str) -> None:
    assert instance_env(instance) == {
        "PATRONICTL_CONFIG_FILE": f"{settings.prefix}/etc/patroni/{pg_version}-test.yaml",
        "PATRONI_NAME": "pg1",
        "PATRONI_SCOPE": "test-scope",
    }


def test_api_request(patroni: models.Patroni, tmp_path: Path) -> None:
    with patch("httpx.request", autospec=True) as request:
        impl.api_request(patroni, "GET", "readiness")
    request.assert_called_once_with(
        "GET",
        "https://localhost:8080/readiness",
        verify=str(tmp_path / "cacert.pem"),
        cert=(str(tmp_path / "host.pem"), str(tmp_path / "host.key")),
    )


def test_check_api_status(settings: Settings, instance: system.Instance) -> None:
    assert settings.patroni
    assert not impl.check_api_status(instance.qualname, settings.patroni)


def test_promote_postgresql(ctx: Context, instance: system.Instance) -> None:
    assert ctx.settings.patroni
    with pytest.raises(exceptions.UnsupportedError):
        ctx.hook.promote_postgresql(ctx=ctx, instance=instance)


@pytest.mark.usefixtures("installed")
def test_cli_patroni_logs(settings: Settings, instance: system.Instance) -> None:
    runner = CliRunner()
    ctx = CLIContext(settings=settings)
    obj = Obj(context=ctx)
    with patch.object(
        impl, "logs", return_value=["l1\n", "l2\n"], autospec=True
    ) as logs:
        result = runner.invoke(cli, ["-i", str(instance), "logs"], obj=obj)
    assert result.exit_code == 0, click_result_traceback(result)
    logs.assert_called_once_with(instance.qualname, ctx.settings.patroni)
    assert result.output == "l1\nl2\n"


def test_preserve_configuration_edits(
    ctx: Context,
    patroni_settings: _patroni.Settings,
    instance: system.Instance,
    instance_manifest: interface.Instance,
) -> None:
    svc = models.ServiceManifest(cluster="test-scope", node="pg1")

    configpath = impl._configpath(instance.qualname, patroni_settings)
    config = yaml.safe_load(configpath.read_text())
    config["bootstrap"]["dcs"]["loop_wait"] = 42

    # Override a managed field already in settings
    config["restapi"]["verify_client"] = "required"
    # Add an unmanaged "extra" field.
    config["restapi"]["http_extra_headers"] = "Custom-Header-Name: Custom Header Value"

    # PostgreSQL settings only defined in actual Patroni configuration file
    # are preserved.
    assert "work_mem" not in config["postgresql"]["parameters"]
    config["postgresql"]["parameters"]["work_mem"] = "16MB"
    config["postgresql"]["pgpass"] = "/home/db/pgpass"
    with configpath.open("w") as f:
        yaml.safe_dump(config, f)
    m = instance_manifest._copy_validate(update={"port": 5467})
    configuration = instances.configuration(ctx, m, instance)
    with impl.setup(
        ctx, instance, m, svc, patroni_settings, configuration, validate=True
    ) as patroni:
        p = patroni.dict()
        assert p["bootstrap"]["dcs"]["loop_wait"] == 42
        assert p["restapi"]["verify_client"] == "required"
        assert (
            p["restapi"]["http_extra_headers"]
            == "Custom-Header-Name: Custom Header Value"
        )
        assert patroni.postgresql.parameters["work_mem"] == "16MB"
        assert patroni.postgresql.parameters["port"] == 5467
        assert patroni.postgresql.listen == "*:5467"
    config = yaml.safe_load(configpath.read_text())
    assert config["postgresql"]["parameters"]["work_mem"] == "16MB"
    assert config["postgresql"]["pgpass"] == "/home/db/pgpass"

    # PostgreSQL settings from instance manifest take precedence over those
    # defined in actual Patroni configuration file.
    configuration["work_mem"] = "42kB"
    with impl.setup(
        ctx, instance, m, svc, patroni_settings, configuration, validate=True
    ) as patroni:
        assert patroni.postgresql.parameters["work_mem"] == "42kB"
    config = yaml.safe_load(configpath.read_text())
    assert config["postgresql"]["parameters"]["work_mem"] == "42kB"



from pathlib import Path
from typing import Any, Literal, Optional

from pydantic import DirectoryPath, Field, FilePath, validator
from pydantic.fields import ModelField

from .. import types
from .base import BaseModel, ConfigPath, LogPath, RunPath


class Etcd(BaseModel):
    """Settings for Etcd (for Patroni)."""

    v2: bool = Field(default=False, description="Configure Patroni to use etcd v2.")

    hosts: tuple[types.Address, ...] = Field(
        default=(types.Address("127.0.0.1:2379"),),
        description="List of etcd endpoint.",
    )

    protocol: Literal["http", "https"] = Field(
        default="http",
        description="http or https, if not specified http is used.",
    )

    cacert: Optional[FilePath] = Field(
        default=None,
        description="Certificate authority to validate the server certificate.",
    )

    cert: Optional[FilePath] = Field(
        default=None,
        description="Client certificate for authentication.",
    )

    key: Optional[FilePath] = Field(
        default=None,
        description="Private key corresponding to the client certificate.",
    )

    @validator("cacert", "cert")
    def __validate_cert_and_protocol_(
        cls, value: Optional[FilePath], values: dict[str, Any]
    ) -> Optional[FilePath]:
        """Make sure protocol https is used when setting certificates."""
        if value is not None and values["protocol"] == "http":
            raise ValueError("'https' protocol is required")
        return value


class WatchDog(BaseModel):
    """Settings for watchdog (for Patroni)."""

    mode: Literal["off", "automatic", "required"] = Field(
        default="off", description="watchdog mode."
    )

    device: Optional[Path] = Field(
        default=None,
        description="Path to watchdog.",
    )

    safety_margin: Optional[int] = Field(
        default=None,
        description=(
            "Number of seconds of safety margin between watchdog triggering"
            " and leader key expiration."
        ),
    )

    @validator("device")
    def __validate_device_(cls, value: Path) -> Path:
        if value and not value.exists():
            raise ValueError(f"path {value} does not exists")
        return value


class RESTAPI(BaseModel):
    """Settings for Patroni's REST API."""

    cafile: Optional[FilePath] = Field(
        default=None,
        description="Certificate authority (or bundle) to verify client certificates.",
    )

    certfile: Optional[FilePath] = Field(
        default=None,
        description="PEM-encoded server certificate to enable HTTPS.",
    )

    keyfile: Optional[FilePath] = Field(
        default=None,
        description="PEM-encoded private key corresponding to the server certificate.",
    )

    verify_client: Optional[Literal["optional", "required"]] = Field(
        default=None, description="Whether to check client certificates."
    )

    @validator("verify_client")
    def __validate_verify_client_and_certfile_(
        cls, value: Optional[Any], values: dict[str, Any]
    ) -> Optional[Any]:
        """Make sure that certfile is set when verify_client is."""
        if value is not None and values.get("certfile") is None:
            raise ValueError("requires 'certfile' to enable TLS")
        return value


class CTL(BaseModel):
    """Settings for Patroni's CTL."""

    certfile: FilePath = Field(
        description="PEM-encoded client certificate.",
    )

    keyfile: FilePath = Field(
        description="PEM-encoded private key corresponding to the client certificate.",
    )


class ServerSSLOptions(BaseModel):
    """Settings for server certificate verification."""

    mode: Optional[
        Literal[
            "disable",
            "allow",
            "prefer",
            "require",
            "verify-ca",
            "verify-full",
        ]
    ] = Field(
        default=None,
        description="Verification mode.",
    )
    crl: Optional[FilePath] = Field(
        default=None,
        description="Certificate Revocation List (CRL).",
    )
    crldir: Optional[DirectoryPath] = Field(
        default=None,
        description="Directory with CRL files.",
    )
    rootcert: Optional[FilePath] = Field(
        default=None,
        description="Root certificate(s).",
    )


class ConnectionOptions(BaseModel):
    ssl: Optional[ServerSSLOptions] = Field(
        default=None,
        description="Settings for server certificate verification when connecting to remote PostgreSQL instances.",
    )


class PostgreSQL(BaseModel):
    connection: Optional[ConnectionOptions] = Field(
        default=None,
        description="Client (libpq) connection options.",
    )
    passfile: ConfigPath = Field(
        default=ConfigPath("patroni/{name}.pgpass"),
        description="Path to .pgpass password file managed by Patroni.",
    )
    use_pg_rewind: bool = Field(
        default=False, description="Whether or not to use pg_rewind."
    )


class Settings(BaseModel):
    """Settings for Patroni."""

    execpath: FilePath = Field(
        default=Path("/usr/bin/patroni"),
        description="Path to patroni executable.",
    )

    configpath: ConfigPath = Field(
        default=ConfigPath("patroni/{name}.yaml"),
        description="Path to the config file.",
    )

    logpath: LogPath = Field(
        default=LogPath("patroni"),
        description="Path where directories are created (based on instance name) to store patroni log files.",
    )

    pid_file: RunPath = Field(
        default=RunPath("patroni/{name}.pid"),
        description="Path to which Patroni process PID will be written.",
    )

    loop_wait: int = Field(
        default=10, description="Number of seconds the loop will sleep."
    )

    etcd: Etcd = Field(default_factory=Etcd, description="Etcd settings.")

    watchdog: WatchDog = Field(
        default_factory=WatchDog, description="Watchdog settings."
    )

    ctl: Optional[CTL] = Field(default=None, description="CTL settings.")

    postgresql: PostgreSQL = Field(
        default_factory=PostgreSQL, description="PostgreSQL settings."
    )

    restapi: RESTAPI = Field(default_factory=RESTAPI, description="REST API settings.")

    @validator("restapi")
    def __validate_restapi_verify_client_(
        cls, value: RESTAPI, values: dict[str, Any], field: ModelField
    ) -> RESTAPI:
        """Make sure 'ctl' client certificates are provided when setting
        restapi.verify_client to required.
        """
        if value.verify_client == "required" and values.get("ctl") is None:
            raise ValueError(
                f"'ctl' must be provided when '{field.name}.verify_client' is set to 'required'"
            )
        return value



from __future__ import annotations

import logging
from typing import Any, Literal

from .. import hookimpl, util
from ..ctx import Context
from ..pm import PluginManager
from ..settings import Settings
from ..types import Status
from . import (
    daemon_reload,
    disable,
    enable,
    get_property,
    install,
    installed,
    restart,
    start,
    stop,
    uninstall,
)

logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    return settings.service_manager == "systemd"


def unit(service: str, qualname: str | None) -> str:
    if qualname is not None:
        return f"pglift-{service}@{qualname}.service"
    else:
        return f"pglift-{service}.service"


@hookimpl
def enable_service(ctx: Context, service: str, name: str | None) -> Literal[True]:
    enable(ctx, unit(service, name))
    return True


@hookimpl
def disable_service(
    ctx: Context, service: str, name: str | None, now: bool | None
) -> Literal[True]:
    kwargs = {}
    if now is not None:
        kwargs["now"] = now
    disable(ctx, unit(service, name), **kwargs)
    return True


@hookimpl
def start_service(ctx: Context, service: str, name: str | None) -> Literal[True]:
    start(ctx, unit(service, name))
    return True


@hookimpl
def stop_service(ctx: Context, service: str, name: str | None) -> Literal[True]:
    stop(ctx, unit(service, name))
    return True


@hookimpl
def restart_service(ctx: Context, service: str, name: str | None) -> Literal[True]:
    restart(ctx, unit(service, name))
    return True


@hookimpl
def service_status(ctx: Context, service: str, name: str | None) -> Status:
    _, status = get_property(ctx, unit(service, name), "ActiveState").split("=", 1)
    status = status.strip()
    return Status.running if status == "active" else Status.not_running


@hookimpl
def site_configure_install(
    settings: Settings, pm: PluginManager, header: str, env: dict[str, Any]
) -> None:
    systemd_settings = settings.systemd
    assert systemd_settings is not None
    for outcome in pm.hook.systemd_unit_templates(
        settings=settings, env=env, content=True
    ):
        for name, content in outcome:
            install(name, util.with_header(content, header), systemd_settings)
    daemon_reload(systemd_settings)


@hookimpl
def site_configure_uninstall(settings: Settings, pm: PluginManager) -> None:
    systemd_settings = settings.systemd
    assert systemd_settings is not None
    for outcome in pm.hook.systemd_unit_templates(
        settings=settings, env=None, content=False
    ):
        for name in outcome:
            uninstall(name, systemd_settings)
    daemon_reload(systemd_settings)


@hookimpl
def site_configure_installed(settings: Settings, pm: PluginManager) -> bool:
    systemd_settings = settings.systemd
    assert systemd_settings is not None
    broken = False
    for outcome in pm.hook.systemd_unit_templates(
        settings=settings, env=None, content=False
    ):
        for name in outcome:
            if not installed(name, systemd_settings):
                logger.error("missing systemd unit '%s'", name)
                broken = True
    return not broken



from pathlib import Path
from typing import Literal, Optional, Union

from pydantic import Field, FilePath

from .base import BaseModel, ConfigPath, DataPath, LogPath, RunPath, ServerCert


class HostRepository(BaseModel):
    """Remote repository host for pgBackRest."""

    host: str = Field(description="Host name of the remote repository.")
    host_port: Optional[int] = Field(
        default=None,
        description="Port to connect to the remote repository.",
    )
    host_config: Optional[Path] = Field(
        default=None,
        description="pgBackRest configuration file path on the remote repository.",
    )


class TLSHostRepository(HostRepository):
    mode: Literal["host-tls"]
    cn: str = Field(description="Certificate Common Name of the remote repository.")
    certificate: ServerCert = Field(
        description="TLS certificate files for the pgBackRest server on site."
    )
    port: int = Field(default=8432, description="Port for the TLS server on site.")
    pid_file: RunPath = Field(
        default=RunPath("pgbackrest.pid"),
        description="Path to which pgbackrest server process PID will be written.",
    )


class SSHHostRepository(HostRepository):
    mode: Literal["host-ssh"]
    host_user: Optional[str] = Field(
        default=None,
        description="Name of the user that will be used for operations on the repository host.",
    )
    cmd_ssh: Optional[Path] = Field(
        default=None,
        description="SSH client command. Use a specific SSH client command when an alternate is desired or the ssh command is not in $PATH.",
    )


class Retention(BaseModel):
    """Retention settings."""

    archive: int = 2
    diff: int = 3
    full: int = 2


class PathRepository(BaseModel):
    """Remote repository (path) for pgBackRest."""

    mode: Literal["path"]
    path: DataPath = Field(
        description="Base directory path where backups and WAL archives are stored.",
    )
    retention: Retention = Field(default=Retention(), description="Retention options.")


class Settings(BaseModel):
    """Settings for pgBackRest."""

    execpath: FilePath = Field(
        default=Path("/usr/bin/pgbackrest"),
        description="Path to the pbBackRest executable.",
    )

    configpath: ConfigPath = Field(
        default=ConfigPath("pgbackrest"),
        description="Base path for pgBackRest configuration files.",
    )

    repository: Union[TLSHostRepository, SSHHostRepository, PathRepository] = Field(
        description="Repository definition, either as a (local) path-repository or as a host-repository.",
        discriminator="mode",
    )

    logpath: LogPath = Field(
        default=LogPath("pgbackrest"),
        description="Path where log files are stored.",
    )

    spoolpath: DataPath = Field(
        default=DataPath("pgbackrest/spool"),
        description="Spool path.",
    )

    lockpath: RunPath = Field(
        default=RunPath("pgbackrest/lock"),
        description="Path where lock files are stored.",
    )



from __future__ import annotations

import configparser
import logging
from collections.abc import Iterator
from pathlib import Path
from typing import ClassVar

import pgtoolkit.conf as pgconf

from .. import cmd, exceptions, hookimpl
from .. import service as service_mod
from .. import systemd, types, util
from ..ctx import Context
from ..models import interface, system
from ..settings import Settings, _pgbackrest, _postgresql
from . import base
from . import register_if as base_register_if
from .base import get_settings, parser

HostRepository = _pgbackrest.TLSHostRepository
logger = logging.getLogger(__name__)


def register_if(settings: Settings) -> bool:
    if not base_register_if(settings):
        return False
    s = get_settings(settings)
    return isinstance(s.repository, HostRepository)


@hookimpl
def site_configure_install(settings: Settings) -> None:
    s = get_settings(settings)
    logger.info("installing pgbackrest server configuration")
    srv_configpath = server_configpath(s)
    srv_configpath.parent.mkdir(parents=True, exist_ok=True)
    config = server_config(s)
    with srv_configpath.open("w") as f:
        config.write(f)

    base.site_configure_install(settings, base_config(s))

    # Also create the log directory here, redundantly with __init__.py,
    # because it's needed when starting the server and we cannot rely on
    # __init__.py hook call as it would happen too late.
    s.logpath.mkdir(exist_ok=True, parents=True)

    srv = Server(s, server_env(settings.postgresql))
    logger.info("starting %s", srv)
    ctx = Context(settings=settings)
    service_mod.start(ctx, srv, foreground=False)
    logger.debug("pinging %s", srv)
    cmd.run(srv.ping_cmd(), check=True)


@hookimpl
def site_configure_uninstall(settings: Settings) -> None:
    s = get_settings(settings)
    srv = Server(s)
    logger.info("stopping %s", srv)
    ctx = Context(settings=settings)
    service_mod.stop(ctx, srv)

    logger.info("uninstalling pgbackrest server configuration")
    server_configpath(s).unlink(missing_ok=True)

    base.site_configure_uninstall(settings)


@hookimpl
def site_configure_uninstalled(settings: Settings) -> None:
    s = get_settings(settings)
    for f in (server_configpath(s), base.config_directory(s), base.base_configpath(s)):
        if not f.exists():
            raise exceptions.InstallationError(
                f"pgBackRest configuration path {f} missing"
            )


SYSTEMD_SERVICE_NAME = "pglift-pgbackrest.service"


@hookimpl
def systemd_unit_templates(
    settings: Settings, content: bool
) -> Iterator[str] | Iterator[tuple[str, str]]:
    if not content:
        yield SYSTEMD_SERVICE_NAME
        return
    s = get_settings(settings)
    yield SYSTEMD_SERVICE_NAME, systemd.template(SYSTEMD_SERVICE_NAME).format(
        executeas=systemd.executeas(settings),
        configpath=server_configpath(s),
        execpath=s.execpath,
        environment=systemd.environment(server_env(settings.postgresql)),
    )


@hookimpl
def instance_configured(
    ctx: Context,
    manifest: interface.Instance,
    config: pgconf.Configuration,
    changes: types.ConfigChanges,
    creating: bool,
    upgrading_from: system.Instance | None,
) -> None:
    with base.instance_configured(
        ctx, manifest, config, changes, creating, upgrading_from
    ) as instance:
        pass

    settings = get_settings(instance._settings)
    srv = Server(settings)
    logger.debug("pinging pgBackRest remote repository %s", srv)
    r = cmd.run(srv.ping_cmd())
    if r.returncode != 0:
        logger.warning("pgBackRest remote repository %s looks unreachable", srv)


@hookimpl
def instance_dropped(ctx: Context, instance: system.Instance) -> None:
    with base.instance_dropped(ctx, instance):
        pass


def server_env(settings: _postgresql.Settings) -> dict[str, str]:
    env = {}
    if settings.auth.passfile is not None:
        env["PGPASSFILE"] = str(settings.auth.passfile)
    return env


class Server:
    """A pgBackRest TLS server, with a Runnable interface."""

    __service_name__: ClassVar = "pgbackrest"
    name: str | None = None

    def __init__(
        self, settings: _pgbackrest.Settings, env: dict[str, str] | None = None
    ) -> None:
        self.settings = settings
        assert isinstance(settings.repository, HostRepository)
        self.repo_settings: HostRepository = settings.repository
        self._env = env

    def __str__(self) -> str:
        return f"pgBackRest TLS server '{self.repo_settings.host}:{self.repo_settings.port}'"

    def args(self) -> list[str]:
        return [
            str(self.settings.execpath),
            "server",
            f"--config={server_configpath(self.settings)}",
        ]

    def pidfile(self) -> Path:
        return Path(str(self.repo_settings.pid_file).format(self.name))

    def env(self) -> dict[str, str] | None:
        return self._env

    def ping_cmd(self, timeout: int = 1) -> list[str]:
        return [
            str(self.settings.execpath),
            "--config=/dev/null",
            "--tls-server-address=*",
            f"--tls-server-port={self.repo_settings.port}",
            "--log-level-file=off",
            "--log-level-console=off",
            "--log-level-stderr=info",
            f"--io-timeout={timeout}",
            "server-ping",
        ]


def repository_settings(settings: _pgbackrest.Settings) -> HostRepository:
    assert isinstance(settings.repository, HostRepository)
    return settings.repository


def server_configpath(settings: _pgbackrest.Settings) -> Path:
    return settings.configpath / "server.conf"


def server_config(settings: _pgbackrest.Settings) -> configparser.ConfigParser:
    """Build the base configuration for the pgbackrest server running on the
    database host.

    This defines the database host as a TLS server, following:
    https://pgbackrest.org/user-guide-rhel.html#repo-host/config
    """
    cp = parser()
    cp.read_string(util.template("pgbackrest", "server.conf").format(**dict(settings)))
    s = repository_settings(settings)
    cp["global"].update(
        {
            "tls-server-address": "*",
            "tls-server-auth": f"{s.cn}=*",
            "tls-server-ca-file": str(s.certificate.ca_cert),
            "tls-server-cert-file": str(s.certificate.cert),
            "tls-server-key-file": str(s.certificate.key),
            "tls-server-port": str(s.port),
        }
    )
    return cp


def base_config(settings: _pgbackrest.Settings) -> configparser.ConfigParser:
    """Build the base configuration for pgbackrest clients on the database
    host.
    """
    cp = parser()
    cp.read_string(
        util.template("pgbackrest", "pgbackrest.conf").format(**dict(settings))
    )
    s = repository_settings(settings)
    rhost = {
        "repo1-host-type": "tls",
        "repo1-host": s.host,
    }
    if s.host_port:
        rhost["repo1-host-port"] = str(s.host_port)
    if s.host_config:
        rhost["repo1-host-config"] = str(s.host_config)
    rhost.update(
        {
            "repo1-host-ca-file": str(s.certificate.ca_cert),
            "repo1-host-cert-file": str(s.certificate.cert),
            "repo1-host-key-file": str(s.certificate.key),
        }
    )
    cp["global"].update(rhost)
    return cp



from __future__ import annotations

import pathlib

from pglift import exceptions


def test_error() -> None:
    err = exceptions.Error("oups")
    assert str(err) == "oups"


def test_notfound() -> None:
    err = exceptions.InstanceNotFound("12/main")
    assert str(err) == "instance '12/main' not found"


def test_configurationerror() -> None:
    err = exceptions.ConfigurationError(
        pathlib.Path("/etc/tool.conf"), "missing 'foo' option"
    )
    assert str(err) == "missing 'foo' option (path: /etc/tool.conf)"



from pydantic import Field

from .base import BaseModel


class Settings(BaseModel):
    """Settings for PoWA."""

    dbname: str = Field(default="powa", description="Name of the PoWA database")
    role: str = Field(default="powa", description="Instance role used for PoWA.")



from __future__ import annotations

import logging

import psycopg.rows
from psycopg import sql

from . import db
from .models import interface

logger = logging.getLogger(__name__)


def ls(cnx: db.Connection) -> list[interface.Publication]:
    with cnx.cursor(row_factory=psycopg.rows.class_row(interface.Publication)) as cur:
        cur.execute(db.query("publications"))
        return cur.fetchall()


def apply(cnx: db.Connection, publication: interface.Publication, dbname: str) -> bool:
    absent = publication.state is interface.PresenceState.absent
    exists = publication.name in {p.name for p in ls(cnx)}
    if not absent and not exists:
        logger.info("creating publication %s in database %s", publication.name, dbname)
        cnx.execute(
            sql.SQL("CREATE PUBLICATION {name} FOR ALL TABLES").format(
                name=sql.Identifier(publication.name)
            )
        )
        return True
    elif absent and exists:
        logger.info(
            "dropping publication %s from database %s", publication.name, dbname
        )
        cnx.execute(
            sql.SQL("DROP PUBLICATION IF EXISTS {name}").format(
                name=sql.Identifier(publication.name)
            )
        )
        return True
    return False



import warnings
from typing import Any

from pydantic import Field, FilePath, validator

from .base import BaseModel, ConfigPath, RunPath


class Settings(BaseModel):
    """Settings for Prometheus postgres_exporter"""

    execpath: FilePath = Field(description="Path to the postgres_exporter executable.")

    role: str = Field(
        default="prometheus",
        description="Name of the PostgreSQL role for Prometheus postgres_exporter.",
    )

    configpath: ConfigPath = Field(
        default=ConfigPath("prometheus/postgres_exporter-{name}.conf"),
        description="Path to the config file.",
    )

    queriespath: ConfigPath = Field(
        default=ConfigPath("prometheus/postgres_exporter_queries-{name}.yaml"),
        description="Path to the queries file.",
    )

    @validator("queriespath")
    def __queriespath_is_deprecated_(cls, value: Any) -> Any:
        warnings.warn(
            "'queriespath' setting is deprecated and will be removed in the next release",
            FutureWarning,
            stacklevel=2,
        )
        return value

    pid_file: RunPath = Field(
        default=RunPath("prometheus/{name}.pid"),
        description="Path to which postgres_exporter process PID will be written.",
    )



from __future__ import annotations

import json
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseSettings, Field, ValidationError

from pglift import exceptions
from pglift.settings import (
    Settings,
    SiteSettings,
    _patroni,
    _pgbackrest,
    _postgresql,
    _systemd,
    base,
)


class SubSubSub(BaseSettings):
    cfg: base.ConfigPath = Field(default=base.ConfigPath("settings.json"))


class SubSub(BaseSettings):
    data: base.DataPath = Field(default=base.DataPath("things"))
    config: SubSubSub = SubSubSub()


class Sub(BaseSettings):
    sub: SubSub
    pid: base.RunPath = Field(default=base.RunPath("pid"))


class S(BaseSettings):
    sub: Sub


def test_prefix_values() -> None:
    bases = {"prefix": Path("/opt"), "run_prefix": Path("/tmp")}
    values = base.prefix_values({"sub": Sub(sub=SubSub())}, bases)
    assert S.parse_obj(values).dict() == {
        "sub": {
            "pid": base.RunPath("/tmp/pid"),
            "sub": {
                "config": {
                    "cfg": Path("/opt/etc/settings.json"),
                },
                "data": Path("/opt/srv/things"),
            },
        },
    }


def test_json_config_settings_source(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    settings = tmp_path / "settings.json"
    pgbindir = tmp_path / "postgres" / "bin"
    pgbindir.mkdir(parents=True)
    json_settings = json.dumps(
        {
            "postgresql": {
                "versions": [{"version": "15", "bindir": str(pgbindir)}],
                "datadir": "/mnt/postgresql/{version}/{name}/data",
            }
        }
    )
    settings.write_text(json_settings)
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", f"@{settings}")
        s = SiteSettings()
    assert s.postgresql.datadir == Path("/mnt/postgresql/{version}/{name}/data")
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", json_settings)
        s = SiteSettings()
    assert s.postgresql.datadir == Path("/mnt/postgresql/{version}/{name}/data")
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", f"@{tmp_path / 'notfound'}")
        with pytest.raises(FileNotFoundError):
            SiteSettings()


def test_yaml_settings(site_settings: MagicMock, tmp_path: Path) -> None:
    bindir = tmp_path / "pgbin"
    bindir.mkdir()
    configdir = tmp_path / "pglift"
    configdir.mkdir()
    settings_fpath = configdir / "settings.yaml"
    settings_fpath.write_text(
        "\n".join(
            [
                "prefix: /tmp",
                "postgresql:",
                "  versions:",
                "    - version: 15",
                f"      bindir: {bindir}",
            ]
        )
    )
    site_settings.return_value = settings_fpath
    s = SiteSettings()
    assert str(s.prefix) == "/tmp"

    settings_fpath.write_text("hello")
    site_settings.return_value = settings_fpath
    with pytest.raises(exceptions.SettingsError, match="expecting an object"):
        SiteSettings()


def test_custom_sources_order(
    site_settings: MagicMock, monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    bindir = tmp_path / "bin"
    bindir.mkdir()
    configdir = tmp_path / "pglift"
    configdir.mkdir()
    settings_fpath = configdir / "settings.yaml"
    settings_fpath.write_text("prefix: /tmp")
    site_settings.return_value = settings_fpath

    with monkeypatch.context() as m:
        m.setenv(
            "SETTINGS",
            json.dumps(
                {
                    "prefix": "/tmp/foo",
                    "postgresql": {
                        "versions": [
                            {"version": 14, "bindir": str(bindir)},
                        ]
                    },
                }
            ),
        )
        s = SiteSettings()
    assert str(s.prefix) == "/tmp/foo"


def test_postgresqlsettings_bindir() -> None:
    with pytest.raises(
        ValidationError, match="missing '{version}' template placeholder"
    ):
        _postgresql.Settings.parse_obj({"bindir": "xxx"})


def test_postgresqlsettings_versions(tmp_path: Path) -> None:
    with pytest.raises(ValidationError, match="either a value is required"):
        _postgresql.Settings.parse_obj({"versions": [], "bindir": None})
    with pytest.raises(ValidationError, match="no value could be inferred"):
        _postgresql.Settings.parse_obj(
            {"versions": [], "bindir": str(tmp_path / "{version}" / "bin")}
        )
    bindir = str(tmp_path / "{version}" / "bin")
    bindir_15 = tmp_path / "15" / "bin"
    bindir_15.mkdir(parents=True)
    s = _postgresql.Settings.parse_obj({"versions": [], "bindir": bindir})
    assert [v.dict() for v in s.versions] == [
        {"bindir": bindir_15, "version": _postgresql.PostgreSQLVersion.v15}
    ]


def test_postgresqlsettings_default_version(tmp_path: Path) -> None:
    with pytest.raises(
        ValidationError, match="value must be amongst declared 'versions': 12"
    ):
        _postgresql.Settings.parse_obj(
            {
                "versions": [{"version": "12", "bindir": str(tmp_path)}],
                "default_version": "13",
                "bindir": str(tmp_path / "{version}" / "bin"),
            }
        )

    with pytest.raises(ValidationError, match="empty 'versions' field"):
        _postgresql.Settings.parse_obj(
            {
                "default_version": "14",
                "bindir": None,
            }
        )
    # Same with a bindir, but unusable to build a 'versions' field.
    with pytest.raises(ValidationError, match="empty 'versions' field"):
        _postgresql.Settings.parse_obj(
            {
                "default_version": "14",
                "bindir": str(tmp_path / "{version}" / "bin"),
            }
        )


def test_role_pgpass(bindir_template: str) -> None:
    base = {"bindir": bindir_template}
    with pytest.raises(
        ValidationError, match="cannot set 'pgpass' without 'auth.passfile'"
    ):
        _postgresql.Settings.parse_obj(
            {"auth": {"passfile": None}, "surole": {"pgpass": True}} | base
        )
    assert not _postgresql.Settings.parse_obj(
        {"auth": {"passfile": None}} | base
    ).surole.pgpass


def test_settings(tmp_path: Path, bindir_template: str) -> None:
    s = Settings(prefix="/", postgresql={"bindir": bindir_template})
    assert hasattr(s, "postgresql")
    assert hasattr(s.postgresql, "datadir")
    assert s.postgresql.datadir == Path("/srv/pgsql/{version}/{name}/data")
    assert s.cli.logpath == Path("/log")

    datadir = tmp_path / "{version}" / "{name}"
    s = Settings.parse_obj(
        {
            "prefix": "/prefix",
            "run_prefix": "/runprefix",
            "postgresql": {"bindir": bindir_template, "datadir": str(datadir)},
        }
    )
    assert s.postgresql.datadir == datadir


def test_settings_nested_prefix(
    tmp_path: Path, pgbackrest_execpath: Path, bindir_template: str
) -> None:
    f = tmp_path / "f"
    f.touch()
    s = Settings.parse_obj(
        {
            "run_prefix": "/test",
            "postgresql": {"bindir": bindir_template},
            "pgbackrest": {
                "execpath": str(pgbackrest_execpath),
                "repository": {
                    "mode": "host-tls",
                    "host": "repo",
                    "cn": "test",
                    "certificate": {"ca_cert": f, "cert": f, "key": f},
                    "pid_file": "backrest.pid",
                },
            },
        }
    )
    assert str(s.dict()["pgbackrest"]["repository"]["pid_file"]) == "/test/backrest.pid"


@pytest.mark.parametrize(
    "obj, expecting",
    [
        ({"postgresql": {"datadir": "/var/lib/{name}"}}, "name, version"),
        (
            {
                "pgbackrest": {
                    "repository": {
                        "mode": "path",
                        "path": "/mnt/backups/{version}-{name}",
                    }
                }
            },
            "none",
        ),
    ],
)
def test_validate_templated_path(
    bindir_template: str, obj: dict[str, Any], expecting: str
) -> None:
    obj.setdefault("postgresql", {})["bindir"] = bindir_template
    with pytest.raises(
        ValidationError,
        match=(
            r"value contains unknown or missing template variable\(s\); expecting: "
            + expecting
        ),
    ):
        Settings.parse_obj(obj)


def test_settings_validate_prefix(postgresql_settings: _postgresql.Settings) -> None:
    with pytest.raises(ValueError, match="expecting an absolute path"):
        Settings(prefix="x", postgresql=postgresql_settings)


def test_settings_validate_service_manager_scheduler(
    postgresql_settings: _postgresql.Settings,
) -> None:
    with pytest.raises(
        ValueError, match="cannot use systemd, if 'systemd' is not enabled globally"
    ):
        Settings(
            service_manager="systemd", postgresql=postgresql_settings
        ).service_manager


@pytest.mark.usefixtures("run_as_root")
def test_settings_as_root() -> None:
    with pytest.raises(
        exceptions.UnsupportedError, match="pglift cannot be used as root"
    ):
        Settings()


def test_postgresql_versions(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    base_bindir = tmp_path / "postgresql"
    base_bindir.mkdir()
    for v in range(13, 16):
        (base_bindir / str(v) / "bin").mkdir(parents=True)
    other_bindir = tmp_path / "pgsql-12" / "bin"
    other_bindir.mkdir(parents=True)
    config: dict[str, Any] = {
        "postgresql": {
            "bindir": str(base_bindir / "{version}" / "bin"),
            "versions": [
                {
                    "version": "12",
                    "bindir": str(other_bindir),
                },
            ],
        },
    }
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(config))
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", f"@{config_path}")
        s = SiteSettings()
    pgversions = s.postgresql.versions
    assert {v.version for v in pgversions} == {"12", "13", "14", "15"}
    assert next(v.bindir for v in pgversions if v.version == "12") == other_bindir
    assert (
        next(v.bindir for v in pgversions if v.version == "13")
        == base_bindir / "13" / "bin"
    )
    config["postgresql"]["default_version"] = "7"
    config_path.write_text(json.dumps(config))
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", f"@{config_path}")
        with pytest.raises(
            ValidationError, match="value is not a valid enumeration member; permitted:"
        ):
            SiteSettings()

    config["postgresql"]["default_version"] = "13"
    config_path.write_text(json.dumps(config))
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", f"@{config_path}")
        s = SiteSettings()
    assert s.postgresql.default_version == "13"

    config["postgresql"]["default_version"] = 7
    config_path.write_text(json.dumps(config))
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", f"@{config_path}")
        with pytest.raises(
            ValidationError, match="value is not a valid enumeration member; permitted:"
        ):
            SiteSettings()

    config["postgresql"]["default_version"] = 13
    config_path.write_text(json.dumps(config))
    with monkeypatch.context() as m:
        m.setenv("SETTINGS", f"@{config_path}")
        s = SiteSettings()
    assert s.postgresql.default_version == "13"


def test_postgresql_dump_restore_commands(bindir_template: str) -> None:
    with pytest.raises(ValidationError) as excinfo:
        _postgresql.Settings.parse_obj(
            {
                "bindir": bindir_template,
                "dump_commands": [
                    ["{bindir}/pg_dump", "--debug"],
                    ["/no/such/file", "{conninfo}"],
                ],
                "restore_commands": [
                    ["not-an-absolute-path", "{dbname}"],
                    ["{bindir}/pg_restore", "ah"],
                ],
            }
        )
    assert excinfo.value.errors() == [
        {
            "loc": ("dump_commands",),
            "msg": "program '/no/such/file' from command #2 does not exist",
            "type": "value_error",
        },
        {
            "loc": ("restore_commands",),
            "msg": "program 'not-an-absolute-path' from command #1 is not an absolute path",
            "type": "value_error",
        },
    ]


def test_systemd_systemctl() -> None:
    with patch("shutil.which", return_value=None, autospec=True) as which:
        with pytest.raises(ValidationError, match="systemctl command not found"):
            _systemd.Settings()
    which.assert_called_once_with("systemctl")


@pytest.mark.usefixtures("systemctl")
def test_systemd_sudo_user() -> None:
    with pytest.raises(ValidationError, match="cannot be used with 'user' mode"):
        _systemd.Settings(user=True, sudo=True)


def test_systemd_disabled(postgresql_settings: _postgresql.Settings) -> None:
    with pytest.raises(ValidationError, match="cannot use systemd"):
        Settings(scheduler="systemd", postgresql=postgresql_settings)
    with pytest.raises(ValidationError, match="cannot use systemd"):
        Settings(service_manager="systemd", postgresql=postgresql_settings)


@pytest.mark.usefixtures("systemctl")
def test_systemd_service_manager_scheduler(
    postgresql_settings: _postgresql.Settings,
) -> None:
    assert (
        Settings(systemd={}, postgresql=postgresql_settings).service_manager
        == "systemd"
    )
    assert (
        Settings(
            systemd={}, service_manager="systemd", postgresql=postgresql_settings
        ).service_manager
        == "systemd"
    )
    assert (
        Settings(
            systemd={}, service_manager=None, postgresql=postgresql_settings
        ).service_manager
        is None
    )


def test_pgbackrest_repository(
    tmp_path: Path, pgbackrest_execpath: Path, bindir_template: str
) -> None:
    f = tmp_path / "f"
    f.touch()
    s = _pgbackrest.Settings.parse_obj(
        {
            "execpath": str(pgbackrest_execpath),
            "repository": {
                "mode": "host-tls",
                "host": "repo",
                "cn": "test",
                "certificate": {"ca_cert": f, "cert": f, "key": f},
            },
        }
    )
    assert isinstance(s.repository, _pgbackrest.HostRepository)

    s = _pgbackrest.Settings.parse_obj(
        {
            "execpath": str(pgbackrest_execpath),
            "repository": {"mode": "path", "path": str(tmp_path)},
        }
    )
    assert isinstance(s.repository, _pgbackrest.PathRepository)

    with pytest.raises(ValidationError, match="repository -> PathRepository -> foo"):
        _pgbackrest.Settings.parse_obj(
            {
                "execpath": str(pgbackrest_execpath),
                "repository": {"mode": "path", "path": str(tmp_path), "foo": 1},
            }
        )
    with pytest.raises(ValidationError, match="repository -> TLSHostRepository -> foo"):
        _pgbackrest.Settings.parse_obj(
            {
                "execpath": str(pgbackrest_execpath),
                "repository": {
                    "mode": "host-tls",
                    "host": "repo",
                    "cn": "test",
                    "certificate": {"ca_cert": f, "cert": f, "key": f},
                    "foo": "bar",
                },
            }
        )


def test_patroni_pgpass(bindir_template: str) -> None:
    with pytest.raises(
        ValidationError,
        match="'patroni.postgresql.passfile' must be different from 'postgresql.auth.passfile'",
    ):
        Settings.parse_obj(
            {
                "postgresql": {
                    "auth": {"passfile": "~/{name}/pgpass"},
                    "bindir": bindir_template,
                },
                "patroni": {"postgresql": {"passfile": "~/{name}/pgpass"}},
            }
        )


def test_patroni_requires_replrole(bindir_template: str) -> None:
    with pytest.raises(
        ValidationError,
        match="'postgresql.replrole' must be provided to use 'patroni'",
    ):
        Settings.parse_obj(
            {
                "postgresql": {
                    "bindir": bindir_template,
                },
                "patroni": {},
            }
        )


def test_patroni_etcd_cert_and_protocol(tmp_path: Path) -> None:
    cacert = tmp_path / "ca.pem"
    cacert.touch()
    with pytest.raises(ValidationError, match="'https' protocol is required"):
        _patroni.Etcd(cacert=cacert)
    _patroni.Etcd(cacert=cacert, protocol="https")
    _patroni.Etcd(protocol="https")


def test_patroni_restapi_verify_client(tmp_path: Path) -> None:
    with pytest.raises(ValidationError, match="requires 'certfile' to enable TLS"):
        _patroni.RESTAPI(verify_client="required")

    certfile = tmp_path / "cert.pem"
    certfile.touch()
    _patroni.RESTAPI(certfile=certfile, verify_client="required")


def test_patroni_restapi_verify_client_ctl(
    bindir_template: str, tmp_path: Path
) -> None:
    certfile = tmp_path / "cert.pem"
    certfile.touch()
    cert = tmp_path / "host.pem"
    cert.touch()
    key = tmp_path / "host.key"
    key.touch()
    with pytest.raises(
        ValidationError,
        match="'ctl' must be provided",
    ):
        _patroni.Settings.parse_obj(
            {
                "restapi": {
                    "certfile": certfile,
                    "verify_client": "required",
                },
            }
        )

    _patroni.Settings.parse_obj(
        {
            "restapi": {
                "certfile": certfile,
                "verify_client": "required",
            },
            "ctl": {
                "certfile": cert,
                "keyfile": key,
            },
        }
    )



from __future__ import annotations

from pglift import logrotate
from pglift.pm import PluginManager
from pglift.settings import Settings


def test_site_configure(settings: Settings, pm: PluginManager) -> None:
    logrotate_settings = settings.logrotate
    assert logrotate_settings is not None
    assert not logrotate_settings.configdir.exists()
    assert not logrotate.site_configure_installed(settings)
    logrotate.site_configure_install(settings, pm)
    assert logrotate.site_configure_installed(settings)
    assert logrotate_settings.configdir.exists()
    config = logrotate_settings.configdir / "logrotate.conf"
    assert config.exists()
    assert settings.pgbackrest is not None
    assert settings.patroni is not None
    assert config.read_text() == "\n".join(
        [
            f"{settings.patroni.logpath}/*/patroni.log {{",
            "  weekly",
            "  rotate 10",
            "  copytruncate",
            "  delaycompress",
            "  compress",
            "  notifempty",
            "  missingok",
            "}",
            "",
            f"{settings.pgbackrest.logpath}/*.log {{",
            "  weekly",
            "  rotate 10",
            "  copytruncate",
            "  delaycompress",
            "  compress",
            "  notifempty",
            "  missingok",
            "}",
            "",
            f"{settings.postgresql.logpath}/*.log {{",
            "  weekly",
            "  rotate 10",
            "  copytruncate",
            "  delaycompress",
            "  compress",
            "  notifempty",
            "  missingok",
            "}",
            "",
        ]
    )
    logrotate.site_configure_uninstall(settings)
    assert not logrotate_settings.configdir.exists()



from __future__ import annotations

import datetime
import fnmatch
import logging
from collections.abc import Iterator
from functools import partial

import attrs
import psycopg
import pytest
from tenacity import Retrying, retry
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_fixed

from pglift import databases, db, exceptions, instances, postgresql
from pglift.ctx import Context
from pglift.models import interface, system
from pglift.settings import Settings
from pglift.settings._postgresql import PostgreSQLVersion

from . import connect, execute
from .conftest import DatabaseFactory, Factory, RoleFactory, TablespaceFactory


@pytest.fixture(scope="module", autouse=True)
def _postgresql_running(instance: system.Instance) -> None:
    if not postgresql.is_running(instance):
        pytest.fail("instance is not running")


@pytest.fixture
def standby_instance_stopped(
    ctx: Context, standby_instance: system.Instance
) -> Iterator[None]:
    instances.stop(ctx, standby_instance)
    try:
        yield
    finally:
        instances.start(ctx, standby_instance)


@pytest.fixture
def pg_database_owner(pg_version: str) -> str:
    return "pg_database_owner" if pg_version >= "14" else "postgres"


def test_exists(
    ctx: Context, instance: system.Instance, database_factory: DatabaseFactory
) -> None:
    assert not databases.exists(ctx, instance, "absent")
    database_factory("present")
    assert databases.exists(ctx, instance, "present")


def test_apply(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
    role_factory: RoleFactory,
    pg_database_owner: str,
) -> None:
    r = execute(
        instance,
        "SELECT default_version FROM pg_available_extensions WHERE name='hstore'",
    )
    assert r is not None
    default_version = r[0]["default_version"]

    dbname = "applyme"
    database = interface.Database(
        name=dbname,
        settings={"work_mem": "1MB"},
        extensions=[{"name": "hstore", "version": default_version}],
        schemas=[{"name": "myapp"}, {"name": "my_schema"}],
        tablespace="pg_default",
    )
    assert not databases.exists(ctx, instance, database.name)

    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.created
    )

    db = databases.get(ctx, instance, dbname)
    assert db.settings == {"work_mem": "1MB"}

    assert db.schemas == [
        interface.Schema(name="my_schema", owner="postgres"),
        interface.Schema(name="myapp", owner="postgres"),
        interface.Schema(name="public", owner=pg_database_owner),
    ]

    assert db.extensions == [
        interface.Extension(name="hstore", schema="public", version=default_version),
    ]

    assert databases.apply(ctx, instance, database).change_state is None  # no-op

    database = interface.Database(
        name=dbname,
        owner="postgres",
        settings={"work_mem": "1MB"},
        schemas=[
            interface.Schema(name="my_schema", owner="postgres"),
            interface.Schema(name="myapp", owner="postgres"),
            interface.Schema(name="public", owner=pg_database_owner),
        ],
        extensions=[
            {"name": "hstore", "schema": "my_schema", "version": default_version},
        ],
        tablespace="pg_default",
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )
    assert databases.get(ctx, instance, dbname) == database

    database = interface.Database(name=dbname, state="absent")
    assert databases.exists(ctx, instance, dbname)
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.dropped
    )
    assert not databases.exists(ctx, instance, dbname)


def test_apply_change_owner(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
    role_factory: RoleFactory,
    pg_database_owner: str,
) -> None:
    database_factory("apply")
    database = interface.Database(name="apply")
    assert databases.apply(ctx, instance, database).change_state is None  # no-op
    assert databases.get(ctx, instance, "apply").owner == "postgres"

    role_factory("dbapply")
    database = interface.Database(
        name="apply",
        owner="dbapply",
        schemas=[interface.Schema(name="public", owner=pg_database_owner)],
        tablespace="pg_default",
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )
    try:
        assert databases.get(ctx, instance, "apply") == database
    finally:
        databases.drop(ctx, instance, interface.DatabaseDropped(name="apply"))


def test_apply_change_tablespace(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
    tablespace_factory: TablespaceFactory,
    standby_instance_stopped: system.Instance,
    pg_database_owner: str,
) -> None:
    database_factory("apply")
    database = interface.Database(name="apply")
    assert databases.apply(ctx, instance, database).change_state is None  # no-op

    tablespace_factory("dbs2")
    database = interface.Database(
        name="apply",
        owner="postgres",
        tablespace="dbs2",
        schemas=[interface.Schema(name="public", owner=pg_database_owner)],
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )
    try:
        assert databases.get(ctx, instance, "apply") == database
    finally:
        databases.drop(ctx, instance, interface.DatabaseDropped(name="apply"))


def test_apply_update_schemas(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
    role_factory: RoleFactory,
    pg_database_owner: str,
) -> None:
    database_factory("db3")
    execute(instance, "CREATE SCHEMA my_schema", fetch=False, dbname="db3")

    assert databases.get(ctx, instance, "db3").schemas == [
        interface.Schema(name="my_schema", owner="postgres"),
        interface.Schema(name="public", owner=pg_database_owner),
    ]

    role_factory("schemauser")
    database = interface.Database(
        name="db3",
        schemas=[
            interface.Schema(name="my_schema", owner="schemauser"),
        ],
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )
    assert databases.get(ctx, instance, "db3").schemas == [
        interface.Schema(name="my_schema", owner="schemauser"),
        interface.Schema(name="public", owner=pg_database_owner),
    ]

    database = interface.Database(
        name="db3",
        schemas=[
            interface.Schema(name="my_schema", state="absent", owner="postgres"),
        ],
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )

    assert databases.get(ctx, instance, "db3").schemas == [
        interface.Schema(name="public", owner=pg_database_owner)
    ]


def test_apply_update_extensions(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
) -> None:
    database_factory("db4")
    execute(
        instance,
        "CREATE SCHEMA my_schema",
        "CREATE SCHEMA second_schema",
        "CREATE EXTENSION unaccent WITH SCHEMA my_schema",
        "CREATE EXTENSION hstore WITH VERSION '1.4'",
        fetch=False,
        dbname="db4",
    )
    r = execute(
        instance,
        "SELECT name, default_version FROM pg_available_extensions",
    )
    assert r is not None
    extversions = {e["name"]: e["default_version"] for e in r}
    unaccent_version = extversions["unaccent"]
    pgss_version = extversions["pg_stat_statements"]
    hstore_version = extversions["hstore"]
    intarray_version = extversions["intarray"]

    assert databases.get(ctx, instance, "db4").extensions == [
        interface.Extension(name="hstore", schema="public", version="1.4"),
        interface.Extension(
            name="unaccent", schema="my_schema", version=unaccent_version
        ),
    ]

    database = interface.Database(
        name="db4",
        extensions=[
            {"name": "pg_stat_statements", "schema": "my_schema"},
            {"name": "unaccent"},
            {"name": "hstore"},
            {"name": "intarray", "version": intarray_version},
        ],
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )
    assert databases.get(ctx, instance, "db4").extensions == [
        interface.Extension(name="hstore", schema="public", version=hstore_version),
        interface.Extension(name="intarray", schema="public", version=intarray_version),
        interface.Extension(
            name="pg_stat_statements", schema="my_schema", version=pgss_version
        ),
        interface.Extension(name="unaccent", schema="public", version=unaccent_version),
    ]

    database = interface.Database(
        name="db4",
        extensions=[
            {"name": "hstore", "state": "absent"},
            {"name": "pg_stat_statements", "state": "absent"},
            {"name": "unaccent", "state": "absent"},
        ],
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )
    assert [
        e.dict(by_alias=True) for e in databases.get(ctx, instance, "db4").extensions
    ] == [{"name": "intarray", "schema": "public", "version": intarray_version}]

    assert databases.apply(ctx, instance, database).change_state is None


def test_publications(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
    pg_database_owner: str,
) -> None:
    database_factory("publisher")
    execute(
        instance,
        "CREATE TABLE things (u int)",
        "CREATE TABLE users (s int)",
        "CREATE TABLE departments (n text)",
        "CREATE PUBLICATION test FOR TABLE users, departments",
        dbname="publisher",
        fetch=False,
    )
    db = databases.get(ctx, instance, "publisher")
    assert db.dict() == {
        "extensions": [],
        "name": "publisher",
        "owner": "postgres",
        "publications": [{"name": "test"}],
        "schemas": [{"name": "public", "owner": pg_database_owner}],
        "settings": None,
        "subscriptions": [],
        "tablespace": "pg_default",
    }

    database = interface.Database(
        name="publisher",
        publications=[{"name": "mypub"}],
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )

    (row,) = execute(
        instance,
        "SELECT puballtables FROM pg_publication WHERE pubname = 'mypub'",
        dbname="publisher",
    )
    assert row["puballtables"] is True

    db = databases.get(ctx, instance, "publisher")
    assert db.dict() == {
        "extensions": [],
        "name": "publisher",
        "owner": "postgres",
        "publications": [
            {"name": "mypub"},
            {"name": "test"},
        ],
        "schemas": [{"name": "public", "owner": pg_database_owner}],
        "settings": None,
        "subscriptions": [],
        "tablespace": "pg_default",
    }

    database = interface.Database(
        name="publisher",
        publications=[{"name": "test", "state": "absent"}],
    )
    assert (
        databases.apply(ctx, instance, database).change_state
        == interface.ApplyChangeState.changed
    )
    db = databases.get(ctx, instance, "publisher")
    assert db.dict()["publications"] == [
        {"name": "mypub"},
    ]

    assert databases.apply(ctx, instance, database).change_state is None


@pytest.fixture
def publisher_role() -> interface.Role:
    return interface.Role.parse_obj(
        {"name": "app", "login": True, "password": "secret", "replication": True}
    )


@pytest.fixture
def published_dbname() -> str:
    return "app"


@pytest.fixture
def publication_name() -> str:
    return "pub"


@pytest.fixture
def publisher_instance(
    settings: Settings,
    publisher_role: interface.Role,
    published_dbname: str,
    publication_name: str,
    surole_password: str | None,
    instance_factory: Factory[tuple[interface.Instance, system.Instance]],
) -> system.Instance:
    _, instance = instance_factory(
        settings.copy(
            update={
                "pgbackrest": None,
                "powa": None,
                "prometheus": None,
                "temboard": None,
            },
            deep=True,
        ),
        name="publisher",
        settings={
            "wal_level": "logical",
            "synchronous_commit": "remote_apply",
        },
        # Either use surole_password which would be in password_command or set
        # a dummy password only needed at instance creation in order to get
        # --auth-local=md5 working.
        surole_password=surole_password or "super",
        auth={
            "local": "md5",
            "host": "md5",
        },
        state="started",
        roles=[publisher_role],
        databases=[
            {
                "name": published_dbname,
                "owner": publisher_role.name,
                "publications": [
                    {"name": publication_name},
                ],
            },
        ],
    )
    assert instance.config()["wal_level"] == "logical"
    execute(
        instance,
        "CREATE TABLE t (s int)",
        dbname=published_dbname,
        role=publisher_role,
        fetch=False,
    )
    return instance


def test_subscriptions(
    ctx: Context,
    publisher_role: interface.Role,
    published_dbname: str,
    publication_name: str,
    publisher_instance: system.Instance,
    role_factory: RoleFactory,
    instance: system.Instance,
    logger: logging.Logger,
    pg_database_owner: str,
) -> None:
    assert publisher_role.password is not None
    publisher_password = publisher_role.password.get_secret_value()
    connection = interface.ConnectionString(
        conninfo=db.dsn(publisher_instance, user=publisher_role.name),
        password=publisher_password,
    )
    with psycopg.connect(connection.full_conninfo):
        # Make sure the connection string to be used for subscription is usable.
        pass
    subname = "subs"
    subscription = {
        "name": subname,
        "connection": connection,
        "publications": [publication_name],
        "enabled": True,
    }
    dbname = "subscriber"
    role_factory(publisher_role.name)
    target = {
        "name": dbname,
        "owner": publisher_role.name,
        "clone": {
            "dsn": f"postgresql://{publisher_role.name}:{publisher_password}@127.0.0.1:{publisher_instance.port}/{published_dbname}",
            "schema_only": True,
        },
        "subscriptions": [subscription],
    }

    @retry(reraise=True, wait=wait_fixed(1), stop=stop_after_attempt(5))
    def check_replication(expected: int) -> None:
        try:
            for line in postgresql.logs(instance, timeout=0):
                logger.debug(line)
        except TimeoutError:
            pass
        for _ in range(3):
            execute(instance, "SELECT pg_sleep(1)", dbname=dbname, fetch=False)
            (row,) = execute(instance, "SELECT MAX(s) AS m FROM t", dbname=dbname)
            if (m := row["m"]) == expected:
                break
        else:
            pytest.fail(f"not replicated: {m} != {expected}")

    try:
        assert (
            databases.apply(
                ctx, instance, interface.Database.parse_obj(target)
            ).change_state
            == interface.ApplyChangeState.created
        )
        actual = databases.get(ctx, instance, dbname)
        assert actual.dict() == {
            "name": dbname,
            "owner": publisher_role.name,
            "settings": None,
            "schemas": [{"name": "public", "owner": pg_database_owner}],
            "extensions": [],
            "publications": [{"name": publication_name}],
            "subscriptions": [
                {
                    "name": "subs",
                    "connection": connection.dict(),
                    "publications": [publication_name],
                    "enabled": True,
                }
            ],
            "tablespace": "pg_default",
        }

        execute(
            publisher_instance,
            "INSERT INTO t VALUES (1), (2)",
            dbname=published_dbname,
            fetch=False,
        )
        check_replication(2)

        subscription["enabled"] = False
        assert (
            databases.apply(
                ctx, instance, interface.Database.parse_obj(target)
            ).change_state
            == interface.ApplyChangeState.changed
        )
        actual = databases.get(ctx, instance, dbname)
        assert actual.dict() == {
            "name": dbname,
            "owner": publisher_role.name,
            "settings": None,
            "schemas": [{"name": "public", "owner": pg_database_owner}],
            "extensions": [],
            "publications": [{"name": publication_name}],
            "subscriptions": [
                {
                    "name": "subs",
                    "connection": connection.dict(),
                    "publications": [publication_name],
                    "enabled": False,
                }
            ],
            "tablespace": "pg_default",
        }

        execute(
            publisher_instance,
            "INSERT INTO t VALUES (10)",
            dbname=published_dbname,
            fetch=False,
        )
        check_replication(2)

    finally:
        if databases.exists(ctx, instance, dbname):
            subscription["state"] = "absent"
            assert (
                databases.apply(
                    ctx, instance, interface.Database.parse_obj(target)
                ).change_state
                == interface.ApplyChangeState.changed
            )
            actual = databases.get(ctx, instance, dbname)
            assert actual.dict()["subscriptions"] == []

            assert (
                databases.apply(
                    ctx, instance, interface.Database.parse_obj(target)
                ).change_state
                is None
            )

            databases.drop(ctx, instance, interface.DatabaseDropped(name=dbname))


@pytest.fixture
def clonable_database(
    ctx: Context,
    role_factory: RoleFactory,
    database_factory: DatabaseFactory,
    instance: system.Instance,
) -> str:
    role_factory("cloner", "LOGIN")
    database_factory("db1", owner="cloner")
    databases.run(
        ctx, instance, "CREATE TABLE persons AS (SELECT 'bob' AS name)", dbnames=["db1"]
    )
    databases.run(ctx, instance, "ALTER TABLE persons OWNER TO cloner", dbnames=["db1"])
    return f"postgresql://cloner@127.0.0.1:{instance.port}/db1"


def test_clone(ctx: Context, clonable_database: str, instance: system.Instance) -> None:
    database = interface.Database.parse_obj(
        {"name": "cloned_db", "clone": {"dsn": clonable_database}}
    )
    assert not databases.exists(ctx, instance, database.name)
    try:
        assert (
            databases.apply(ctx, instance, database).change_state
            == interface.ApplyChangeState.created
        )
        result = execute(instance, "SELECT * FROM persons", dbname="cloned_db")
        assert result == [{"name": "bob"}]
    finally:
        databases.drop(ctx, instance, interface.DatabaseDropped(name="cloned_db"))

    database = interface.Database.parse_obj(
        {
            "name": "cloned_schema",
            "clone": {"dsn": clonable_database, "schema_only": True},
        }
    )
    assert database.clone and database.clone.schema_only
    assert not databases.exists(ctx, instance, database.name)
    try:
        assert (
            databases.apply(ctx, instance, database).change_state
            == interface.ApplyChangeState.created
        )
        result = execute(instance, "SELECT * FROM persons", dbname="cloned_schema")
        assert result == []
    finally:
        databases.drop(ctx, instance, interface.DatabaseDropped(name="cloned_schema"))

    # DSN which target is a non existing database
    options = interface.CloneOptions(
        dsn=f"postgresql://postgres@127.0.0.1:{instance.port}/nosuchdb"
    )
    with pytest.raises(exceptions.CommandError) as cm:
        databases.clone("cloned", options, instance)
    assert cm.value.cmd[0] == str(instance.bindir / "pg_dump")
    assert not databases.exists(ctx, instance, "cloned")

    # DSN which target is a non existing user
    options = interface.CloneOptions(
        dsn=f"postgresql://nosuchuser@127.0.0.1:{instance.port}/postgres"
    )
    with pytest.raises(exceptions.CommandError) as cm:
        databases.clone("cloned", options, instance)
    assert cm.value.cmd[0] == str(instance.bindir / "pg_dump")
    assert not databases.exists(ctx, instance, "cloned")

    # Target database does not exist
    with pytest.raises(exceptions.CommandError) as cm:
        databases.clone("nosuchdb", database.clone, instance)
    assert cm.value.cmd[0] == str(instance.bindir / "pg_restore")
    assert not databases.exists(ctx, instance, "nosuchdb")


def test_get(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
    pg_database_owner: str,
) -> None:
    with pytest.raises(exceptions.DatabaseNotFound, match="absent"):
        databases.get(ctx, instance, "absent")

    database_factory("describeme")
    execute(instance, "ALTER DATABASE describeme SET work_mem TO '3MB'", fetch=False)
    execute(
        instance,
        "CREATE SCHEMA my_schema",
        "CREATE EXTENSION unaccent WITH SCHEMA my_schema",
        fetch=False,
        dbname="describeme",
    )
    database = databases.get(ctx, instance, "describeme")
    assert database.name == "describeme"
    assert database.settings == {"work_mem": "3MB"}
    assert database.schemas == [
        interface.Schema(name="my_schema", owner="postgres"),
        interface.Schema(name="public", owner=pg_database_owner),
    ]
    r = execute(
        instance,
        "SELECT default_version FROM pg_available_extensions WHERE name='unaccent'",
    )
    assert r is not None
    default_version = r[0]["default_version"]
    assert database.extensions == [
        interface.Extension(
            name="unaccent", schema="my_schema", version=default_version
        )
    ]


def test_encoding(instance: system.Instance) -> None:
    with connect(instance) as conn:
        assert databases.encoding(conn) == "UTF8"


def test_ls(
    ctx: Context, instance: system.Instance, database_factory: DatabaseFactory
) -> None:
    database_factory("db1")
    database_factory("db2")
    dbs = databases.ls(ctx, instance)
    dbnames = [d.name for d in dbs]
    assert "db2" in dbnames
    dbs = databases.ls(ctx, instance, dbnames=("db1",))
    dbnames = [d.name for d in dbs]
    assert "db2" not in dbnames
    assert len(dbs) == 1
    db1 = attrs.asdict(next(d for d in dbs))
    db1.pop("size")
    db1["tablespace"].pop("size")
    assert db1 == {
        "acls": [],
        "collation": "C",
        "ctype": "C",
        "description": None,
        "encoding": "UTF8",
        "name": "db1",
        "owner": "postgres",
        "tablespace": {"location": "", "name": "pg_default"},
    }


def test_drop(
    ctx: Context, instance: system.Instance, database_factory: DatabaseFactory
) -> None:
    with pytest.raises(exceptions.DatabaseNotFound, match="absent"):
        databases.drop(ctx, instance, interface.DatabaseDropped(name="absent"))

    database_factory("dropme")
    databases.drop(ctx, instance, interface.DatabaseDropped(name="dropme"))
    assert not databases.exists(ctx, instance, "dropme")


def test_drop_force(
    ctx: Context,
    pg_version: str,
    instance: system.Instance,
    database_factory: DatabaseFactory,
) -> None:
    database_factory("dropme")

    if pg_version >= PostgreSQLVersion.v13:
        with connect(instance, dbname="dropme"):
            with pytest.raises(psycopg.errors.ObjectInUse):
                databases.drop(ctx, instance, interface.DatabaseDropped(name="dropme"))
            databases.drop(
                ctx, instance, interface.DatabaseDropped(name="dropme", force_drop=True)
            )
        assert not databases.exists(ctx, instance, "dropme")
    else:
        with pytest.raises(
            exceptions.UnsupportedError,
            match=r"^Force drop option can't be used with PostgreSQL < 13$",
        ):
            databases.drop(
                ctx, instance, interface.DatabaseDropped(name="dropme", force_drop=True)
            )


def test_run(
    ctx: Context,
    instance: system.Instance,
    database_factory: DatabaseFactory,
    caplog: pytest.LogCaptureFixture,
) -> None:
    database_factory("test")
    caplog.clear()
    with caplog.at_level(logging.INFO, logger="pglift"):
        result_run = databases.run(
            ctx,
            instance,
            "CREATE TABLE persons AS (SELECT 'bob' AS name)",
            dbnames=["test"],
        )
    assert "CREATE TABLE persons AS (SELECT 'bob' AS name)" in caplog.records[0].message
    assert "SELECT 1" in caplog.records[1].message
    assert not result_run
    result = execute(instance, "SELECT * FROM persons", dbname="test")
    assert result == [{"name": "bob"}]
    result_run = databases.run(
        ctx,
        instance,
        "SELECT * from persons",
        dbnames=["test"],
    )
    assert result_run == {"test": [{"name": "bob"}]}


def test_run_analyze(
    ctx: Context, instance: system.Instance, database_factory: DatabaseFactory
) -> None:
    database_factory("test")

    @retry(
        reraise=True,
        retry=retry_if_exception_type(AssertionError),
        stop=stop_after_attempt(5),
        wait=wait_fixed(1),
    )
    def last_analyze() -> datetime.datetime:
        result = execute(
            instance,
            "SELECT MIN(last_analyze) m FROM pg_stat_all_tables WHERE last_analyze IS NOT NULL",
            dbname="test",
        )[0]["m"]
        assert isinstance(result, datetime.datetime), result
        return result

    retrying = partial(
        Retrying,
        retry=retry_if_exception_type(AssertionError),
        stop=stop_after_attempt(5),
        wait=wait_fixed(0.2),
        reraise=True,
    )

    databases.run(ctx, instance, "ANALYZE")
    previous = last_analyze()
    databases.run(ctx, instance, "ANALYZE")
    for attempt in retrying():
        now = last_analyze()
        with attempt:
            assert now > previous
    databases.run(ctx, instance, "ANALYZE", exclude_dbnames=["test"])
    for attempt in retrying():
        with attempt:
            assert last_analyze() == now


def test_run_output_notices(
    ctx: Context, instance: system.Instance, capsys: pytest.CaptureFixture[str]
) -> None:
    databases.run(
        ctx, instance, "DO $$ BEGIN RAISE NOTICE 'foo'; END $$", dbnames=["postgres"]
    )
    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == "foo\n"


def test_dump(
    ctx: Context, instance: system.Instance, database_factory: DatabaseFactory
) -> None:
    with pytest.raises(exceptions.DatabaseNotFound, match="absent"):
        databases.dump(ctx, instance, "absent")
    database_factory("dbtodump")
    databases.dump(ctx, instance, "dbtodump")
    directory = instance.dumps_directory
    assert directory.exists()
    (dumpfile,) = list(directory.iterdir())
    assert fnmatch.fnmatch(str(dumpfile), "*dbtodump_*.dump"), dumpfile



from __future__ import annotations

from pathlib import Path

import pytest
from pgtoolkit import conf as pgconf

from pglift import conf
from pglift.settings import Settings


def test_read(pg_version: str, settings: Settings, tmp_path: Path) -> None:
    datadir = tmp_path
    postgresql_auto_conf = datadir / "postgresql.auto.conf"
    postgresql_auto_conf.write_text("primary_conn_info = host=primary\n")
    postgresql_conf = datadir / "postgresql.conf"
    postgresql_conf.write_text("\n".join(["bonjour = hello", "port=1234"]))

    config = conf.read(datadir)
    assert config.bonjour == "hello"
    assert config.port == 1234
    assert config.primary_conn_info == "host=primary"

    config = conf.read(datadir, managed_only=True)
    assert config.bonjour == "hello"
    assert config.port == 1234
    assert "primary_conn_info" not in config

    postgresql_conf.unlink()
    with pytest.raises(FileNotFoundError, match=str(postgresql_conf)):
        conf.read(datadir, True)


def test_update(datadir: Path, write_changes: bool) -> None:
    cfg = pgconf.parse(datadir / "postgresql.conf.sample")
    conf.update(
        cfg,
        max_connections=10,  # changed
        bonjour=True,  # uncommented
        log_destination="stderr",  # added
    )
    fpath = datadir / "postgresql.conf"
    if write_changes:
        cfg.save(fpath)
    expected = fpath.read_text().splitlines(keepends=True)
    assert cfg.lines == expected


def test_changes() -> None:
    assert conf.changes(
        {"unchanged": "x", "changed": 5432, "removed": "rmme"},
        {"unchanged": "x", "changed": 5433, "added": "x,y,z"},
    ) == {
        "changed": (5432, 5433),
        "removed": ("rmme", None),
        "added": (None, "x,y,z"),
    }



from __future__ import annotations

import sys
from functools import cache
from importlib import metadata
from typing import Any, Callable, Final, TypeVar, overload

import pluggy

from . import pm, settings

__all__ = ["hookimpl"]

# Declare type for hookimpl on our side until a version (> 1.0.0) is
# available.

F = TypeVar("F", bound=Callable[..., Any])


@overload
def hookimpl(__func: F) -> F:
    ...


@overload
def hookimpl(**kwargs: Any) -> Callable[[F], F]:
    ...


def hookimpl(*args: Any, **kwargs: Any) -> Any:
    return pluggy.HookimplMarker(__name__)(*args, **kwargs)


def version() -> str:
    return metadata.version(__name__)


@cache
def plugin_manager(s: settings.Settings) -> pm.PluginManager:
    return pm.PluginManager.get(s)


execpath: Final = (
    sys.executable
    if getattr(sys, "frozen", False)
    else f"{sys.executable} -m {__name__}"
)



from __future__ import annotations

from pathlib import Path

import pytest
from pydantic import SecretStr

from pglift import passfile as passfile_mod
from pglift.models import interface
from pglift.models.system import Instance
from pglift.settings import Settings


class Role(interface.Role):
    def __init__(
        self, name: str, password: str | None = None, pgpass: bool = False
    ) -> None:
        super().__init__(
            name=name,
            password=SecretStr(password) if password is not None else None,
            pgpass=pgpass,
        )


@pytest.fixture
def passfile(settings: Settings) -> Path:
    fpath = settings.postgresql.auth.passfile
    assert fpath is not None
    fpath.write_text("*:999:*:edgar:fbi\n")
    return fpath


@pytest.mark.parametrize(
    "role, changed, pgpass",
    [
        (Role("alice"), False, "*:999:*:edgar:fbi\n"),
        (Role("bob", "secret"), False, "*:999:*:edgar:fbi\n"),
        (Role("charles", pgpass=True), False, "*:999:*:edgar:fbi\n"),
        (Role("danny", "sss", True), True, "*:999:*:danny:sss\n*:999:*:edgar:fbi\n"),
        (Role("edgar", "cia", True), True, "*:999:*:edgar:cia\n"),
        (Role("edgar", None, False), True, ""),
    ],
)
def test_role_change(
    instance: Instance, passfile: Path, role: Role, changed: bool, pgpass: str
) -> None:
    assert passfile_mod.role_change(instance=instance, role=role) == changed
    assert passfile.read_text() == pgpass


def test_role_inspect(instance: Instance) -> None:
    fpath = instance._settings.postgresql.auth.passfile
    assert fpath is not None
    fpath.write_text("*:999:*:edgar:fbi\n")
    assert passfile_mod.role_inspect(instance, "edgar") == {"pgpass": True}
    assert passfile_mod.role_inspect(instance, "alice") == {"pgpass": False}



from __future__ import annotations

import attrs
import pytest
from psycopg.conninfo import conninfo_to_dict

from pglift import exceptions
from pglift.ctx import Context
from pglift.models import system
from pglift.models.system import Instance
from pglift.settings import Settings
from pglift.settings._postgresql import PostgreSQLVersion


def test_baseinstance(instance: Instance) -> None:
    assert isinstance(instance.version, PostgreSQLVersion)


def test_baseinstance_str(pg_version: str, instance: Instance) -> None:
    assert str(instance) == f"{pg_version}/test"


def test_baseinstance_qualname(pg_version: str, instance: Instance) -> None:
    assert instance.qualname == f"{pg_version}-test"


@pytest.mark.parametrize(
    ["attrname", "expected_suffix"],
    [
        ("datadir", "srv/pgsql/{version}/test/data"),
        ("waldir", "srv/pgsql/{version}/test/wal"),
    ],
)
def test_baseinstance_paths(
    pg_version: str, instance: Instance, attrname: str, expected_suffix: str
) -> None:
    path = getattr(instance, attrname)
    assert path.match(expected_suffix.format(version=pg_version))


def test_baseinstance_get(settings: Settings, pg_version: str) -> None:
    i = system.BaseInstance.get("test", None, settings)
    assert i.version == pg_version


def test_baseinstance_get_invalid_version(settings: Settings) -> None:
    if any(
        v.bindir.exists() for v in settings.postgresql.versions if v.version == "12"
    ):
        pytest.skip("PostgreSQL 12 installed")
    with pytest.raises(
        exceptions.InvalidVersion, match="version 12 unsupported in site settings"
    ):
        system.BaseInstance.get("onze", "12", settings)


def test_postgresqlinstance_system_lookup(
    settings: Settings, instance: Instance
) -> None:
    i = system.PostgreSQLInstance.system_lookup(instance)
    expected = system.PostgreSQLInstance(instance.name, instance.version, settings)
    assert i == expected

    i = system.PostgreSQLInstance.system_lookup(
        (instance.name, instance.version, settings)
    )
    assert i == expected

    with pytest.raises(TypeError, match="expecting either a BaseInstance or"):
        system.PostgreSQLInstance.system_lookup(("nameonly",))  # type: ignore[arg-type]


def test_instance_validate(settings: Settings, pg_version: str) -> None:
    class Service:
        pass

    with pytest.raises(
        ValueError, match="values for 'services' field must be of distinct types"
    ):
        system.Instance(
            name="invalid",
            version=pg_version,
            settings=settings,
            services=[Service(), Service()],
        )

    class Service2:
        pass

    i = system.Instance(
        name="valid",
        version=pg_version,
        settings=settings,
        services=[Service(), Service2()],
    )
    assert i.services


def test_instance_system_lookup(settings: Settings, instance: Instance) -> None:
    i = system.Instance.system_lookup(instance)
    assert i == instance

    i = system.Instance.system_lookup((instance.name, instance.version, settings))
    assert i == instance


def test_instance_system_lookup_misconfigured(instance: Instance) -> None:
    (instance.datadir / "postgresql.conf").unlink()
    with pytest.raises(exceptions.InstanceNotFound, match=str(instance)):
        system.Instance.system_lookup(instance)


def test_postgresqlinstance_exists(pg_version: str, settings: Settings) -> None:
    instance = system.PostgreSQLInstance(
        name="exists", version=pg_version, settings=settings
    )
    with pytest.raises(exceptions.InstanceNotFound):
        instance.exists()
    instance.datadir.mkdir(parents=True)
    (instance.datadir / "PG_VERSION").write_text(pg_version)
    with pytest.raises(exceptions.InstanceNotFound):
        instance.exists()
    (instance.datadir / "postgresql.conf").touch()
    assert instance.exists()


def test_postgresqlinstance_port(instance: Instance) -> None:
    assert instance.port == 999


def test_postgresqlinstance_socket_directory(instance: Instance) -> None:
    assert instance.socket_directory == "/socks"

    (instance.datadir / "postgresql.conf").write_text(
        "unix_socket_directories = '@a , b '\n"
    )
    assert instance.socket_directory == "b"

    (instance.datadir / "postgresql.conf").write_text("pif = paf\n")
    assert instance.socket_directory is None


def test_postgresqlinstance_config(instance: Instance) -> None:
    assert instance.config().as_dict() == {
        "port": 999,
        "unix_socket_directories": "/socks, /shoes",
    }


def test_postgresqlinstance_standby_for(
    ctx: Context, instance: Instance, standby_instance: Instance
) -> None:
    assert not instance.standby
    assert standby_instance.standby
    assert conninfo_to_dict(standby_instance.standby.primary_conninfo) == {
        "host": "/tmp",
        "port": "4242",
        "user": "pg",
    }
    assert standby_instance.standby.slot == "aslot"


def test_privileges_sorted() -> None:
    p = system.Privilege(
        database="postgres",
        schema="main",
        object_type="table",
        object_name="foo",
        role="postgres",
        privileges=["select", "delete", "update"],
        column_privileges={"postgres": ["update", "delete", "reference"]},
    )
    assert attrs.asdict(p) == {
        "column_privileges": {"postgres": ["delete", "reference", "update"]},
        "database": "postgres",
        "object_name": "foo",
        "object_type": "table",
        "privileges": ["delete", "select", "update"],
        "role": "postgres",
        "schema": "main",
    }



from __future__ import annotations

import importlib
from typing import Any

import pluggy

from . import __name__ as pkgname
from . import hookspecs, settings
from ._compat import Self


class PluginManager(pluggy.PluginManager):
    modules = (
        "postgresql",
        "databases",
        "passfile",
        "backup",
        "logrotate",
        "pgbackrest",
        "patroni",
        "pgbackrest.repo_host_tls",
        "pgbackrest.repo_host_ssh",
        "pgbackrest.repo_path",
        "prometheus",
        "powa",
        "rsyslog",
        "temboard",
        "systemd.service_manager",
        "systemd.scheduler",
    )

    @classmethod
    def all(cls) -> Self:
        """Return a PluginManager with all modules registered."""
        self = cls(pkgname)
        self.add_hookspecs(hookspecs)
        for hname in cls.modules:
            hm = importlib.import_module(f"{pkgname}.{hname}")
            self.register(hm)
        return self

    @classmethod
    def get(cls, settings: settings.Settings) -> Self:
        """Return a PluginManager based on 'settings'."""
        self = cls(pkgname)
        self.add_hookspecs(hookspecs)
        for hname in cls.modules:
            hm = importlib.import_module(f"{pkgname}.{hname}")
            if not hasattr(hm, "register_if") or hm.register_if(settings):
                self.register(hm)
        return self

    def register(self, plugin: Any, name: str | None = None) -> Any:
        rv = super().register(plugin, name)
        assert self.get_hookcallers(plugin), f"{plugin} has no hook caller"
        return rv

    def unregister_all(self) -> list[object]:
        unregistered = []
        for __, plugin in self.list_name_plugin():
            self.unregister(plugin)
            unregistered.append(plugin)
        return unregistered

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.get_plugins() == other.get_plugins()
